import pytest
from conftest import (
    base_path,
    btl_path,
    cnv_path,
    hex_path,
    proc_template,
    test_cnv,
)

from ctdam.parser import (
    BottleFile,
    CnvCollection,
    CnvFile,
    DataFile,
    FileCollection,
    HexCollection,
    XMLCONFile,
    get_collection,
)
from ctdam.parser.casts import Casts


@pytest.mark.parametrize(
    ("path", "suffix", "file_type", "num_of_files"),
    [
        (btl_path, ".btl", BottleFile, 1),
    ],
)
class TestCollections:
    @pytest.fixture
    def files(
        self,
        path: str,
        suffix: str,
        file_type: DataFile,
        num_of_files: int,
    ):
        files = get_collection(
            path_to_files=path,
            file_suffix=suffix,
        )
        return files

    @pytest.fixture
    def df(self, files: FileCollection):
        return files.get_collection_dataframe()

    def test_collected_dataframe(self, files: FileCollection):
        assert (
            files.df.size
            == sum([file.df.shape[0] for file in files])
            * files.df_list[0].shape[1]
        )

    def test_cast_specific_info_added(
        self,
        files: FileCollection,
        num_of_files: int,
    ):
        df = files.get_collection_dataframe(
            files.get_dataframes(event_log=True, coordinates=True)
        )

        assert len(df.Event.unique()) == num_of_files
        assert len(df.Longitude.unique()) == num_of_files
        assert len(df.Latitude.unique()) == num_of_files

    def test_cnv_specifics(self, files: FileCollection):
        if files.file_suffix != "cnv":
            pytest.skip("cnv-specific test")
        assert files.data_meta_info["c0mS/cm"]["name"] == "Conductivity"
        df = files.tidy_collection_dataframe(files.df)
        assert df.isna().sum().sum() == 0
        assert files.array.shape == (
            sum([file.parameters.full_data_array.shape[0] for file in files]),
            files[0].parameters.full_data_array.shape[1],
        )
        assert files.array.shape == files.df.shape
        assert files.get_dataframes(time_correction=True)


def test_mixed_processing_steps():
    cnvs = CnvCollection(cnv_path, only_metadata=True)
    with pytest.warns(UserWarning):
        cnvs.get_processing_steps()


class TestHexCollections:
    @pytest.fixture
    def files(self) -> FileCollection:
        files = HexCollection(
            path_to_files=hex_path,
            pattern="SO308-2",
            sorting_key=lambda file: int(file.stem.split("_")[1]),
        )
        return files

    def test_pattern_handling(self, files):
        assert len(files) == 3

    def test_xmlcon_matching(self, files):
        xmlcon = XMLCONFile(hex_path.joinpath("SO308-2.XMLCON"))
        for file in files:
            assert file.xmlcon == xmlcon


@pytest.mark.long
@pytest.mark.parametrize(
    ("data_path", "pattern", "size"),
    [
        # test cnvs
        (cnv_path, "SO308", 3),
        # test hexes and file type detection
        (base_path, "", 11),
        # test file type detection with pattern
        (base_path, "cnv", 9),
        # test path to single file
        (hex_path.joinpath("MSM138_10-1.hex"), "", 1),
    ],
)
class TestCasts:
    @pytest.fixture
    def files(self, data_path, pattern) -> Casts:
        return Casts(
            path_to_data=data_path,
            processing_info=proc_template,
            pattern=pattern,
        )

    def test_base(self, files, size, tmp_path):
        assert len(files) + len(files.anomalous_data) == size
        files.to_tsv(tmp_path.joinpath(files.cruise))

    def test_sensor_info(self, files, size):
        if size < 4:
            files.read_sensor_info()
            assert len(files.sensor_info) == 1
        else:
            with pytest.warns():
                files.read_sensor_info()
            assert len(files.sensor_info) > size * 0.5


def test_ctddata_in_casts(tmp_path):
    cnv = CnvFile(test_cnv)
    casts = Casts(
        ctd_data=[cnv.to_ctd_data()],
        plot=True,
        show_plot=False,
        plot_dir=tmp_path,
    )
    assert tmp_path.joinpath(test_cnv.name).with_suffix(".html").exists()
    casts.read_sensor_info()
    assert len(casts.sensor_info) == 1



import json
import xml.etree.ElementTree as ET
from collections import UserDict
from pathlib import Path

import xmltodict

from ctdam.utils import UnexpectedFileFormat


class XMLFile(UserDict):
    """
    Parent class for XML and psa representation that loads XML as a
    python-internal tree and as a dict.

    Parameters
    ----------
    path_to_file : Path | str :
        the path to the xml file

    Returns
    -------

    """

    def __init__(self, path_to_file: Path | str):
        self.path_to_file = Path(path_to_file)
        self.file_name = self.path_to_file.name
        self.file_dir = self.path_to_file.parents[0]
        self.input = ""
        with open(self.path_to_file, "r") as file:
            for line in file:
                self.input += line
        self.xml_tree = ET.fromstring(self.input)
        self.data = xmltodict.parse(self.input)

    def __eq__(self, other) -> bool:
        """
        Allows comparison of two instances of this class.

        Uses the parsed xml information to determine equality.

        Parameters
        ----------
        other: XMLFile
            An instance of this class.

        Returns
        -------
        Whether the given instance and this one are equal.
        """
        return self.data == other.data

    def to_xml(self, file_name=None, file_path=None):
        """
        Writes the dictionary to xml.

        Parameters
        ----------
        file_name : str :
            the original files name (Default value = self.file_name)
        file_path : pathlib.Path :
            the directory of the file (Default value = self.file_dir)

        Returns
        -------

        """
        file_path = self.file_dir if file_path is None else file_path
        file_name = self.file_name if file_name is None else file_name
        with open(Path(file_path).joinpath(file_name), "w") as file:
            file.write(xmltodict.unparse(self.data, pretty=True))

    def to_json(self, file_name=None, file_path=None):
        """
        Writes the dictionary representation of the XML input to a json
        file.

        Parameters
        ----------
        file_name : str :
            the original files name (Default value = self.file_name)
        file_path : pathlib.Path :
            the directory of the file (Default value = self.file_dir)

        Returns
        -------

        """
        file_path = self.file_dir if file_path is None else file_path
        file_name = self.file_name if file_name is None else file_name
        with open(Path(file_path).joinpath(file_name + ".json"), "w") as file:
            json.dump(self.data, file, indent=4)


class XMLCONFile(XMLFile):
    """ """

    def __init__(self, path_to_file):
        super().__init__(path_to_file)
        self.sensor_info = self.get_sensor_info()

    def get_sensor_info(self) -> list[dict]:
        """
        Creates a multilevel dictionary, dropping the first four dictionaries,
        to retrieve pure sensor information.

        Returns
        -------
        A list of all the individual sensor information, stored in dictionaries

        """
        try:
            sensors = self.data["SBE_InstrumentConfiguration"]["Instrument"][
                "SensorArray"
            ]["Sensor"]
        except KeyError as error:
            raise UnexpectedFileFormat("XMLCON", error)
        else:
            # create a tidied version of the xml-parsed sensor dict
            sensor_names = []
            tidied_sensor_list = []
            for entry in sensors:
                sensor_key = list(entry.keys())[-1]
                if not sensor_key.endswith(("Sensor", "Meter")):
                    continue
                sensor_name = sensor_key.removesuffix("Sensor")
                # the wetlab sensors feature a suffix _Sensor
                sensor_name = sensor_name.removesuffix("_")
                # assuming, that the first sensor in the xmlcon is also on the
                # first sensor strand, the second occurence of the name is
                # suffixed with '2'
                if sensor_name in sensor_names:
                    sensor_name += "2"
                sensor_names.append(sensor_name)
                # move the calibration info one dictionary level up
                calibration_info = entry[sensor_key]
                # build the new dictionary
                try:
                    new_dict = {
                        "Channel": str(int(entry["@index"]) + 1),
                        "SensorName": sensor_name,
                        **calibration_info,
                    }
                except TypeError:
                    new_dict = {
                        "Channel": entry["@Channel"],
                        "SensorName": sensor_name,
                        "Info": calibration_info,
                    }
                tidied_sensor_list.append(new_dict)
            return tidied_sensor_list


class PsaFile(XMLFile):
    """ """

    def __init__(self, path_to_file):
        super().__init__(path_to_file)



import logging
from datetime import datetime
from pathlib import Path

import numpy as np
import pandas as pd
import xmltodict

from ctdam.utils import (
    create_event_string,
    extract_sensor_name,
    read_event_name,
)

logger = logging.getLogger(__name__)


class DataFile:
    """
    The base class for all Sea-Bird data files, which are .cnv, .btl, and .bl .
    One instance of this class, or its children, represents one data text file.
    The different information bits of such a file are structured into individual
    lists or dictionaries. The data table will be loaded as numpy array and
    can be converted to a pandas DataFrame. Datatype-specific behavior is
    implemented in the subclasses.


    Parameters
    ----------
    path_to_file: Path | str :
        The file to the data file.
    only_header: bool :
        Whether to stop reading the file after the metadata header.
    """

    def __init__(
        self,
        path_to_file: Path | str,
        only_header: bool = False,
    ):
        self.path_to_file = Path(path_to_file)
        self.file_name = self.path_to_file.stem
        self.file_dir = self.path_to_file.parent
        self.only_header = only_header
        self.raw_file_data = []  # the text file input
        self.header = []  # the full file header
        self.sbe9_data = []  # device specific information
        self.metadata = {}  # non-SeaBird metadata
        self.metadata_list = []  # unstructured metadata for easier export
        self.data_table_description = []  # the column names and other info
        self.sensor_data = []
        self.sensors = {}  # xml-parsed sensor data
        self.processing_info = []  # everything after the sensor data
        self.data = []  # the data table
        self.file_data = self.raw_file_data  # variable file information
        self.read_file()
        self.metadata = self.structure_metadata(self.metadata_list)
        if len(self.sensor_data) > 0:
            self.sensors = self.sensor_xml_to_flattened_dict(
                "".join(self.sensor_data)
            )
        self.start_time = self.reading_start_time()
        self.read_event_information()

    def __str__(self) -> str:
        return "/n".join(self.file_data)

    def __repr__(self) -> str:
        return str(self.path_to_file.absolute())

    def __eq__(self, other) -> bool:
        return self.file_data == other.file_data

    def read_event_information(
        self,
        regex_string: str = r"(?P<c>[a-z]{1,3}\d{1,3})(-|_|\/)?(?P<cn>1|2)?(-|_)(?P<s>\d{1,4})(-|_)(?P<e>\d{1,2})",
        leading_zeroes: bool = False,
    ):
        self.cruise, self.station = read_event_name(
            self.file_name,
            regex_string,
        )
        if "Station" in self.metadata and not self.station:
            station_string = self.metadata["Station"]
            self.cruise, self.station = read_event_name(
                station_string,
                regex_string,
            )
        self.event_name = create_event_string(
            self.cruise,
            self.station,
            leading_zeroes,
        )
        self.metadata["Station"] = self.event_name

    def read_file(self):
        """
        Reads and structures all the different information present in the
        file. Lists and Dictionaries are the data structures of choice. Uses
        basic prefix checking to distinguish different header information.
        """
        past_bad_flag = False
        with self.path_to_file.open("r", encoding="latin-1") as file:
            for line in file:
                self.raw_file_data.append(line)
                if line.startswith("* "):
                    self.header.append(line)
                    self.sbe9_data.append(line[2:])
                elif line.startswith("**"):
                    self.header.append(line)
                    self.metadata_list.append(line[3:])
                elif line.startswith("#"):
                    self.header.append(line)
                    if line[2:].strip().startswith("<"):
                        self.sensor_data.append(line[2:])
                    else:
                        if past_bad_flag:
                            self.processing_info.append(line[2:])
                        else:
                            self.data_table_description.append(line[2:])
                            if line.startswith("# bad_flag"):
                                past_bad_flag = True
                else:
                    if line.startswith("*END*"):
                        self.header.append(line)
                        if self.only_header:
                            break
                    else:
                        self.data.append(line)

    def reading_start_time(self) -> datetime | None:
        """
        Extracts the Cast start time from the metadata header.
        """
        start_time = None
        for line in self.header:
            if line.startswith(("* System UTC", "* NMEA UTC")):
                start_time = line.split("=", 1)[1].strip()
                break
        if start_time:
            start_time = datetime.strptime(start_time, "%b %d %Y %H:%M:%S")
        return start_time

    def sensor_xml_to_flattened_dict(
        self, sensor_data: str
    ) -> list[dict] | dict:
        """
        Reads the pure xml sensor input and creates a multilevel dictionary,
        dropping the first two dictionaries, as they are single entry only

        Parameters
        ----------
        sensor_data: str:
            The raw xml sensor data.

        Returns
        -------
        A list of sensor information, which is a structured dict.

        """
        full_sensor_dict = xmltodict.parse(sensor_data, process_comments=True)
        try:
            sensors = full_sensor_dict["Sensors"]["sensor"]
        except KeyError as error:
            logger.error(f"XML is not formatted as expected: {error}")
            return full_sensor_dict
        else:
            # create a tidied version of the xml-parsed sensor dict
            return extract_sensor_name(sensors)

    def structure_metadata(self, metadata_list: list) -> dict:
        """
        Creates a dictionary to store custom metadata, of which Sea-Bird allows
        12 lines in each file.

        Parameters
        ----------
        metadata_list: list :
            a list of the individual lines of metadata found in the file

        Returns
        -------
        a dictionary of the lines of metadata divided into key-value pairs
        """
        out_dict = {}
        for line in metadata_list:
            if "=" in line:
                (key, val) = line.split("=")
            elif ":" in line:
                (key, val) = line.split(":")
            else:
                key = line
                val = ""
            out_dict[key.strip()] = val.strip()
        return out_dict

    def define_output_path(
        self,
        file_path: Path | str | None = None,
        file_name: str | None = None,
        file_type: str = ".csv",
    ) -> Path:
        """
        Creates a Path object holding the desired output path.

        Parameters
        ----------
        file_path : Path :
            directory the file sits in (Default value = self.file_dir)
        file_name : str :
            the original file name (Default value = self.file_name)
        file_type : str :
            the output file type (Default = '.csv')
        Returns
        -------
        a Path object consisting of the full path of the new file

        """
        file_path = self.file_dir if file_path is None else file_path
        file_name = self.file_name if file_name is None else file_name
        if file_type[0] != ".":
            file_type = "." + file_type
        return Path(file_path).joinpath(file_name).with_suffix(file_type)

    def to_csv(
        self,
        data: pd.DataFrame | np.ndarray,
        with_header: bool = True,
        output_file_path: Path | str | None = None,
        output_file_name: str | None = None,
    ):
        """
        Writes a csv from the given data.

        Parameters
        ----------
        data: pd.DataFrame | np.ndarray :
            The source data to use.
        with_header : boolean :
            indicating whether the header shall appear in the output
             (Default value = True)
        output_file_path : Path :
            file directory (Default value = None)
        output_file_name : str :
            original file name (Default value = None)

        Returns
        -------

        """
        new_file_path = self.define_output_path(
            output_file_path, output_file_name
        )
        if with_header:
            with open(new_file_path, "w") as file:
                for line in self.header:
                    file.write(line)
        if isinstance(data, pd.DataFrame):
            data.to_csv(new_file_path, index=False, mode="a")
        else:
            np.savetxt(new_file_path, data, delimiter=",")

    def selecting_columns(
        self,
        list_of_columns: list | str,
        df: pd.DataFrame,
    ):
        """
        Alters the dataframe to only hold the given columns.

        Parameters
        ----------
        list_of_columns: list or str : a collection of columns
        df : pandas.Dataframe :
            Dataframe (Default value = None)

        Returns
        -------

        """
        # ensure that the input is a list, so that isin() can do its job
        if isinstance(list_of_columns, str):
            list_of_columns = [list_of_columns]
        if isinstance(df, pd.DataFrame):
            self.df = df[list_of_columns].reset_index(drop=True)



from .datafiles import *
from .bottlefile import *
from .bottlelogfile import *
from .parameter import *
from .processing_steps import *
from .cnvfile import *
from .xmlfiles import *
from .hexfile import *
from .ctddata import *
from .geomar_ctd_file_parser import *
from .file_collection import *



import logging
from datetime import datetime, time
from pathlib import Path
from typing import Union

import numpy as np
import pandas as pd

from ctdam.parser import DataFile

logger = logging.getLogger(__name__)


class BottleFile(DataFile):
    """Class that represents a SeaBird Bottle File. Organizes the files table
    information into a pandas dataframe. This allows the usage of this
    powerful library for statistics, visualization, data manipulation, export,
    etc.

    Parameters
    ----------

    Returns
    -------

    """

    def __init__(self, path_to_file: Path | str, only_header: bool = False):
        super().__init__(path_to_file, only_header)
        if not only_header:
            self.original_df = self.create_dataframe()
            self.df = self.original_df
            self.setting_dataframe_dtypes()
            self.adding_timestamp_column()

    def create_dataframe(self):
        """Creates a dataframe out of the btl file. Manages the double data
        header correctly.

        Parameters
        ----------

        Returns
        -------

        """
        # TODO: this needs to be broken down into smaller pieces...
        top_names, bottom_names = self.reading_data_header()
        # creating statistics column to store the row type information:
        # 4 rows per bottle, average, standard deviation, max value, min value
        top_names.append("Statistic")
        # TODO: sexier way to construct dataframe than opening the file a
        # second time
        df: pd.DataFrame = pd.read_fwf(
            self.path_to_file,
            index_col=False,
            skiprows=len(self.header) + 2,
            header=None,
            names=top_names,
        )

        # handling the double row header
        rowtypes = df[df.columns[-1]].unique()

        # TODO: can this be made a little pretier?
        def separate_double_header_row(df, column, length):
            """

            Parameters
            ----------
            df :
            column :
            length :

            Returns
            -------

            """
            column_idx = df.columns.get_loc(column)
            old_column = df.iloc[::length, column_idx].reset_index(drop=True)
            new_column = df.iloc[1::length, column_idx].reset_index(drop=True)
            old_column_expanded = pd.Series(
                np.repeat(old_column, length)
            ).reset_index(drop=True)
            new_column_expanded = pd.Series(
                np.repeat(new_column, length)
            ).reset_index(drop=True)
            df[column] = old_column_expanded
            df.insert(
                column_idx + 1, bottom_names[column_idx], new_column_expanded
            )
            return df

        df = separate_double_header_row(df, "Date", len(rowtypes))
        df = separate_double_header_row(df, top_names[0], len(rowtypes))
        # remove brackets around statistics values
        df["Statistic"] = df["Statistic"].str.strip("()")
        df = df.rename(
            mapper={"Btl_ID": "Bottle_ID", "Bottle": "Bottle_ID"}, axis=1
        )
        return df

    def adding_timestamp_column(self):
        """Creates a timestamp column that holds both, Date and Time
        information.

        Parameters
        ----------

        Returns
        -------

        """
        # constructing timestamp column
        timestamp = []
        for datepoint, timepoint in zip(self.df.Date, self.df.Time):
            timestamp.append(
                datetime.combine(datepoint, time.fromisoformat(str(timepoint)))
            )
        self.df.insert(2, "Timestamp", timestamp)
        self.df.Timestamp = pd.to_datetime(self.df.Timestamp)

    def setting_dataframe_dtypes(self):
        """Sets the types for the column values in the dataframe."""
        # setting dtypes
        # TODO: extending this to the other columns!
        self.df.Date = pd.to_datetime(self.df.Date)
        self.df.Bottle_ID = self.df.Bottle_ID.astype(int)

    def selecting_rows(
        self, df=None, statistic_of_interest: Union[list, str] = ["avg"]
    ):
        """Creates a dataframe with the given row identifier, using the
        statistics column. A single string or a list of strings can be
        processed.

        Parameters
        ----------
        df : pandas.Dataframe :
            the files Pandas representation (Default value = self.df)
        statistic_of_interest: list or str :
            collection of values of the 'statistics' column in self.df
             (Default value = ['avg'])

        Returns
        -------

        """
        df = self.df if df is None else df
        # ensure that the input is a list, so that isin() can do its job
        if isinstance(statistic_of_interest, str):
            statistic_of_interest = [statistic_of_interest]
        self.df = df.loc[df["Statistic"].isin(statistic_of_interest)]

    def reading_data_header(self):
        """Identifies and separatly collects the rows that specify the data
        tables headers.

        Parameters
        ----------

        Returns
        -------

        """
        n = 11  # fix column width of a seabird btl file
        top_line = self.data[0]
        second_line = self.data[1]
        top_names = [
            top_line[i : i + n].split()[0]
            for i in range(0, len(top_line) - n, n)
        ]
        bottom_names = [
            second_line[i : i + n].split()[0] for i in range(0, 2 * n, n)
        ]
        return top_names, bottom_names

    def add_station_and_event_column(self):
        event_list = [self.metadata["Station"] for _ in self.data]
        self.df.insert(0, "Event", pd.Series(event_list))

    def add_position_columns(self):
        latitude_list = [self.metadata["GPS_Lat"] for _ in self.data]
        self.df.insert(1, "Latitude", pd.Series(latitude_list))
        longitude_list = [self.metadata["GPS_Lon"] for _ in self.data]
        self.df.insert(2, "Longitude", pd.Series(longitude_list))



import logging
import multiprocessing
from pathlib import Path

from ctdam.proc.seabird.module import ProcessingModule
from ctdam.proc.seabird.step import ProcessingStep
from ctdam.proc.settings import IncompleteConfigFile
from ctdam.proc.utils import fill_file_type_dir

logger = logging.getLogger(__name__)


class ProcessingRoutine:
    """
    Batch processing of running multiple processing modules in sequence is
    realised by this class. It also allows batch processing of multiple files
    at the same time.
    In order to accomplish that, a very specific input is needed, specified in
    detail by a configuration .toml file. For that reason, a thorough input
    check takes place upon initialization.

    Parameters
    ----------
    processing_info: dict :
        All the information necessary to run a processing routine. Can be built
        internally in other python code or, more often, will be the input from
        a .toml configuration file.

    """

    def __init__(self, processing_info: dict):
        self.processing_info = processing_info
        try:
            self.exe_dir = Path(processing_info["exe_directory"])
            self.psa_dir = Path(processing_info["psa_directory"])
        except KeyError:
            error_message = f"The input {
                processing_info
            } is missing an exe_dir and/or psa_dir, you need to specify both of them."
            raise IncompleteConfigFile(error_message)
        try:
            self.file_list = [
                Path(file) for file in processing_info["file_list"]
            ]
            assert len(self.file_list) > 0
        except (KeyError, AssertionError):
            try:
                self.input_dir = Path(processing_info["input_directory"])
            except KeyError:
                error_message = f"The input {
                    processing_info
                } is missing an input_dir and/or a file name, specify one of them."
                raise IncompleteConfigFile(error_message)
        try:
            self.xmlcons = Path(processing_info["xmlcons"])
            assert len(str(self.xmlcons)) > 1
        except (KeyError, AssertionError):
            self.xmlcons = None
        try:
            self.output_dir = Path(processing_info["output_directory"])
        except KeyError:
            try:
                self.output_dir = self.input_dir.parent
            except KeyError:
                self.output_dir = self.file_list[0].parent
        try:
            self.file_type_dir = Path(processing_info["file_type_directory"])
        except KeyError:
            self.file_type_dir = None
        try:
            self.verbose = processing_info["verbose"]
        except KeyError:
            self.verbose = False
        self.module_info = {
            key: value for key, value in processing_info["modules"].items()
        }
        if list(self.module_info.keys())[0] == "datcnv":
            self.input_data = "hex"
        else:
            self.input_data = "cnv"
        self.modules = self.create_modules()
        self.steps = self.create_steps()

    def create_modules(self) -> list[ProcessingModule]:
        """
        Loads the module information into individual module instances.

        A module is the representation of one Sea-Bird module of the same name.
        It stores the path to the executable and a configuration .psa file.
        """
        out_list = []
        for name, info in self.module_info.items():
            try:
                psa_path = info["psa"]
            except KeyError:
                psa_path = None
            try:
                exe_path = info["exe"]
            except KeyError:
                exe_path = None
            try:
                new_file_suffix = info["file_suffix"]
            except KeyError:
                new_file_suffix = ""
            out_list.append(
                ProcessingModule(
                    name=name,
                    exe_dir=self.exe_dir,
                    psa_dir=self.psa_dir,
                    psa_path=psa_path,
                    exe_path=exe_path,
                    new_file_suffix=new_file_suffix,
                )
            )
        return out_list

    def create_steps(self) -> list[ProcessingStep]:
        """
        Creates a list of processing steps from the input information.

        A step is the application of one processing module to a target data
        file, which usually is either .hex or .cnv .
        """
        # TODO: break into smaller pieces
        out_list = []
        try:
            files = self.file_list
        except AttributeError:
            files = [
                file for file in self.input_dir.rglob(f"*{self.input_data}")
            ]
        for file in files:
            file = Path(file)
            input_dir = file.parent
            if self.input_data == "hex":
                xmlcon = self.find_xmlcon(file)
                if not xmlcon:
                    continue
            else:
                xmlcon = None
            for index, module in enumerate(self.modules):
                if index == 0:
                    input_path = file
                else:
                    input_path = self.output_dir.joinpath(
                        file.with_suffix(".cnv").name
                    )
                if module.name == "bottlesum":
                    input_path = input_dir.joinpath(
                        file.with_suffix(".ros").name
                    )
                    if not input_path.exists():
                        continue
                    new_file_name = file.with_suffix(".btl").name
                elif module.name == "iow_btl_id":
                    input_path = input_path.joinpath(
                        file.with_suffix(".btl").name
                    )
                    if not input_path.exists():
                        continue
                    new_file_name = file.with_suffix(".btl").name
                elif module.name == "derive":
                    xmlcon = self.find_xmlcon(file)
                    new_file_name = file.with_suffix(".cnv").name
                else:
                    new_file_name = file.with_suffix(".cnv").name
                out_list.append(
                    ProcessingStep(
                        module=module,
                        input_path=input_path,
                        xmlcon_path=xmlcon,
                        output_path=self.output_dir,
                        new_name=str(new_file_name),
                        verbose=self.verbose,
                    )
                )
            if self.file_type_dir:
                for target_file in input_dir.iterdir():
                    if target_file.stem == file.stem:
                        fill_file_type_dir(self.file_type_dir, target_file)
        return out_list

    def find_xmlcon(self, input_file: Path) -> Path | None:
        """
        Recursively searches for a corresponding .XMLCON file to a given input
        file.

        Parameters
        ----------
        input_file: Path :
            The path to the target input file.


        Returns
        -------
        A path to the .XMLCON file, or none, if none found.

        """
        if self.xmlcons:
            parent_dir = self.xmlcons
        else:
            parent_dir = input_file.parent
        for xmlcon in parent_dir.rglob("*.XMLCON", case_sensitive=False):
            if xmlcon.stem.lower() == input_file.stem.lower():
                return xmlcon
        logger.warning(
            f'Could not find a matching XMLCON for the file "{
                input_file.stem
            }" inside of "{parent_dir}".'
        )

    def run(self, timeout: int = 60):
        """
        Performs the processing by calling the steps run methods in sequence.
        """
        for step in self.steps:
            step.run(timeout=timeout)

    def start_thread(self):
        """
        Starts a separate process to run the processing.

        This leverages parallel computing capabilities and allows the targeted
        killing of the processing, when necessary.
        """
        self.run_thread = multiprocessing.Process(target=self.run)
        self.run_thread.start()

    def cancel(self):
        """Kills the running processing process."""
        logger.info(f"Interupted processing of \n{self.processing_info}")
        self.run_thread.kill()



import logging
from pathlib import Path

import pytest

base_path = Path("sbs_data")
cnv_path = base_path.joinpath("cnv")
hex_path = base_path.joinpath("hex")
psa_path = base_path.joinpath("psa")
btl_path = base_path.joinpath("btl")
test_hex = hex_path.joinpath("EMB356_11-1.hex")
test_cnv = cnv_path.joinpath("EMB356_11-1.cnv")
proc_template = {
    "input": "",
    "output_type": "internal",
    "modules": {
        "wildedit_geomar": {},
        "wfilter": {},
        "alignctd": {},
        "celltm": {},
        "binavg": {},
    },
}

logger = logging.getLogger(__name__)


def pytest_addoption(parser):
    """
    Adds a command line option to pytest that controls the skipping of code
    that runs seabird processing modules.
    """
    parser.addoption(
        "--run_seabird",
        "-S",
        action="store_true",
        default=False,
        help="Whether to run seabird processing modules during the tests.",
    )
    parser.addoption(
        "--run_long",
        "-L",
        action="store_true",
        default=False,
        help="Whether to run tests for the long cnvs as well.",
    )
    parser.addoption(
        "--create_files",
        "-F",
        action="store_true",
        default=False,
        help="Whether to write cnv files to disk.",
    )


@pytest.fixture
def run_seabird_modules(request) -> bool:
    """
    Makes the boolean flag of the command line option available to individual
    tests.
    """
    return request.config.getoption("--run_seabird")


@pytest.fixture
def run_long_tests(request) -> bool:
    """
    Makes the boolean flag of the command line option available to individual
    tests.
    """
    return request.config.getoption("--run_long")


@pytest.fixture
def create_files(request) -> bool:
    """
    Makes the boolean flag of the command line option available to individual
    tests.
    """
    return request.config.getoption("--create_files")


def pytest_collection_modifyitems(config, items):
    skip_non_seabird = pytest.mark.skip(reason="No seabird option given.")
    skip_long = pytest.mark.skip(reason="Test too long.")
    for item in items:
        if not config.getoption("--run_seabird"):
            if "seabird" in item.keywords:
                item.add_marker(skip_non_seabird)
        if not config.getoption("--run_long"):
            if "long" in item.keywords:
                item.add_marker(skip_long)


def check_and_remove_file(output_file: Path | str):
    output_file = Path(output_file)
    if output_file.exists():
        assert True
        output_file.unlink()
    else:
        logger.error(f"Could not find the expected file: {output_file}")
        assert False



import logging
from datetime import timedelta
from pathlib import Path

import numpy as np
import pandas as pd

from ctdam.parser import CnvProcessingSteps, DataFile, Parameters

logger = logging.getLogger(__name__)


class CnvFile(DataFile):
    """
    A representation of a cnv-file as used by SeaBird.

    This class intends to fully extract and organize the different types of
    data and metadata present inside of such a file. Downstream libraries shall
    be able to use this representation for all applications concerning cnv
    files, like data processing, transformation or visualization.

    To achieve that, the metadata header is organized by the parent-class,
    DataFile, while the data table is extracted by this class. The data
    representation can be a numpy array or pandas dataframe. The handling of
    the data is mostly done inside parameters, a representation of the
    individual measurement parameter data and metadata.

    This class is also able to parse the edited data and metadata back to the
    original .cnv file format, allowing for custom data processing using this
    representation, while still being able to use Sea-Birds original software
    on that output. It also allows to stay comparable with other parsers or
    methods in general.

    Parameters
    ----------
    path_to_file: Path | str:
        the path to the file
    only_header: bool :
        Whether to stop reading the file after the metadata header.
    create_dataframe: bool :
        Whether to create a pandas DataFrame from the data table.
    absolute_time_calculation: bool:
        whether to use a real timestamp instead of the second count
    event_log_column: bool:
        whether to add a station and device event column from DSHIP
    coordinate_columns: bool:
        whether to add longitude and latitude from the extra metadata header

    """

    def __init__(
        self,
        path_to_file: Path | str,
        only_header: bool = False,
        create_dataframe: bool = False,
        absolute_time_calculation: bool = False,
        event_log_column: bool = False,
        coordinate_columns: bool = False,
    ):
        super().__init__(path_to_file, only_header)
        self.processing_steps = self.get_processing_step_infos()
        self.parameters = Parameters(
            self.data, self.data_table_description, only_header
        )
        if create_dataframe:
            self.df = self.create_dataframe()
        if absolute_time_calculation:
            self.absolute_time_calculation()
        if event_log_column:
            self.add_station_and_event_column()
        if coordinate_columns:
            self.add_position_columns()

    def create_dataframe(self) -> pd.DataFrame:
        """
        Plain dataframe creator.
        """
        self.df = self.parameters.get_pandas_dataframe()
        return self.df

    def absolute_time_calculation(self) -> bool:
        """
        Replaces the basic cnv time representation of counting relative to the
        casts start point, by real UTC timestamps.
        This operation will act directly on the dataframe.

        """
        time_parameter = None
        for parameter in self.parameters.keys():
            if parameter.lower().startswith("time"):
                time_parameter = parameter
        if time_parameter and self.start_time:
            self.parameters.create_parameter(
                name="datetime",
                data=np.array(
                    [
                        timedelta(days=float(time)) + self.start_time
                        if time_parameter == "timeJ"
                        else timedelta(seconds=float(time)) + self.start_time
                        for time in self.parameters[time_parameter].data
                    ],
                    dtype=str,
                ),
            )
            return True
        return False

    def add_start_time(self) -> bool:
        """
        Adds the Cast start time to the dataframe.
        Necessary for joins on the time.
        """
        if self.start_time:
            self.parameters.create_parameter(
                name="start_time",
                data=str(self.start_time),
            )
            return True
        return False

    def get_processing_step_infos(self) -> CnvProcessingSteps:
        """
        Collects the individual validation modules and their respective
        information, usually present in key-value pairs.
        """
        return CnvProcessingSteps(self.processing_info)

    def df2cnv(self, df: pd.DataFrame | None = None) -> list:
        """
        Parses a pandas dataframe into a list that represents the lines inside
        of a cnv data table.

        Parameters
        ----------
        df: DataFrame to export, default is self.df

        Returns
        -------
        a list of lines in the cnv data table format

        """
        df = df if isinstance(df, pd.DataFrame) else self.df
        cnv_out = []
        for _, row in df.iterrows():
            cnv_like_row = "".join(
                (lambda column: f"{str(column):>11}")(value) for value in row
            )
            cnv_out.append(cnv_like_row + "\r\n")
        return cnv_out

    def array2cnv(self) -> list:
        result = []
        for row in self.parameters.full_data_array:
            formatted_row = "".join(str(elem).rjust(11) for elem in row)
            result.append(formatted_row + "\r\n")
        return result

    def to_cnv(
        self,
        file_name: Path | str | None = None,
        use_dataframe: bool = False,
    ):
        """
        Writes the values inside of this instance as a new cnv file to disc.

        Parameters
        ----------
        file_name: Path:
            the new file name to use for writing
        use_current_df: bool:
            whether to use the current dataframe as data table
        use_current_validation_header: bool:
            whether to use the current processing module list
        header_list: list:
            the data columns to use for the export

        """
        file_name = self.path_to_file if file_name is None else file_name
        # content construction
        if use_dataframe:
            data = self.df2cnv()
        else:
            data = self.array2cnv()
        self._update_header()
        self.file_data = [*self.header, *data]
        # writing content out
        try:
            with open(file_name, "w", encoding="latin-1") as file:
                for line in self.file_data:
                    file.write(line)

        except IOError as error:
            logger.error(f"Could not write cnv file: {error}")

    def to_ctd_data(self):
        from ctdam.parser.ctddata import CTDData

        ctd_data = CTDData(
            parameters=self.parameters,
            metadata_source=self,
        )

        return ctd_data

    def _update_header(self):
        """Re-creates the cnv header."""
        self.data_table_description = self.parameters._form_data_table_info()
        self.processing_info = self.processing_steps._form_processing_info()
        self.header = [
            *[f"* {data}" for data in self.sbe9_data[:-1]],
            *[
                f"** {key} = {value}\r\n" if value else f"** {key}\r\n"
                for key, value in self.metadata.items()
            ],
            f"* {self.sbe9_data[-1]}",
            *[f"# {data}" for data in self.data_table_description],
            *[f"# {data}" for data in self.sensor_data],
            *[f"# {data}" for data in self.processing_info],
            "*END*\r\n",
        ]
        self.data = self.array2cnv()
        self.file_data = [*self.header, *self.data]

    def add_processing_metadata(self, module: str, key: str, value: str):
        """
        Adds new processing lines to the list of processing module information

        Parameters
        ----------
        module: str :
            the name of the processing module
        key: str :
            the description of the value
        value: str :
            the information

        """
        self.processing_steps.add_info(module, key, value)
        self._update_header()

    def add_station_and_event_column(self) -> bool:
        """
        Adds a column with the DSHIP station and device event numbers to the
        dataframe. These must be present inside the extra metadata header.

        """
        if "Station" in self.metadata:
            data_value = self.metadata["Station"]
            return_value = True
        else:
            data_value = None
            return_value = False
        self.parameters.create_parameter(
            data=data_value,
            name="Event",
        )
        return return_value

    def add_position_columns(self) -> bool:
        """
        Adds a column with the longitude and latitude to the dataframe.
        These must be present inside the extra metadata header.

        """
        if ("latitude" or "longitude") in [
            k.lower() for k in self.parameters.keys()
        ]:
            return True
        if ("GPS_Lat" and "GPS_Lon") in self.metadata:
            self.parameters.create_parameter(
                data=self.metadata["GPS_Lat"],
                name="Latitude",
            )
            self.parameters.create_parameter(
                data=self.metadata["GPS_Lon"],
                name="Longitude",
            )
            return True
        else:
            return False

    def add_cast_number(self, number: int | None = None) -> bool:
        """
        Adds a column with the cast number to the dataframe.

        Parameters
        ----------
        number: int:
            the cast number of this files cast

        """
        if ("Cast" in self.metadata.keys()) and (not number):
            number = int(self.metadata["Cast"])
        if number:
            self.parameters.create_parameter(
                data=number,
                name="Cast",
            )
            return True
        return False



import importlib.metadata
import logging
import sys
from datetime import datetime, timezone
from pathlib import Path

import numpy as np
import pandas as pd
from seabirdscientific import instrument_data as sbs_id

from ctdam.conv.cast_borders import get_cast_borders
from ctdam.conv.parameter_mapping import ParameterMapping
from ctdam.parser.ctddata import CTDData
from ctdam.parser.hexfile import HexFile
from ctdam.parser.parameter import Parameters
from ctdam.parser.processing_steps import CnvProcessingSteps
from ctdam.parser.xmlfiles import XMLCONFile

logger = logging.getLogger(__name__)


def hex_reading(hex: HexFile) -> pd.DataFrame:
    instrument_info = hex.xmlcon["SBE_InstrumentConfiguration"]["Instrument"]
    instrument_name = instrument_info["Name"]
    # TODO: extend this
    device_mapping = {"SBE 911plus": sbs_id.InstrumentType.SBE911Plus}
    try:
        for device in device_mapping.keys():
            if instrument_name.startswith(device):
                instrument_type = device_mapping[device]
        assert instrument_type

    except Exception:
        sys.exit(f"Unknown instrument: {instrument_name}. Aborting.")
    enabled_sensors = []

    sensor_info = hex.xmlcon.sensor_info
    sensor_names = [s["SensorName"] for s in sensor_info]
    if "Temperature" in sensor_names:
        enabled_sensors.append(sbs_id.Sensors.Temperature)
    if "Conductivity" in sensor_names:
        enabled_sensors.append(sbs_id.Sensors.Conductivity)
    if "Pressure" in sensor_names:
        enabled_sensors.append(sbs_id.Sensors.Pressure)
    if "Temperature2" in sensor_names:
        enabled_sensors.append(sbs_id.Sensors.SecondaryTemperature)
    if "Conductivity2" in sensor_names:
        enabled_sensors.append(sbs_id.Sensors.SecondaryConductivity)

    voltage_sensors = [
        sbs_id.Sensors.ExtVolt0,
        sbs_id.Sensors.ExtVolt1,
        sbs_id.Sensors.ExtVolt2,
        sbs_id.Sensors.ExtVolt3,
        sbs_id.Sensors.ExtVolt4,
        sbs_id.Sensors.ExtVolt5,
        sbs_id.Sensors.ExtVolt6,
        sbs_id.Sensors.ExtVolt7,
    ]
    enabled_sensors = [*enabled_sensors, *voltage_sensors]

    if instrument_info["SurfaceParVoltageAdded"] == "1" and [
        l for l in hex.sbe9_data if l.startswith("surface PAR")
    ]:
        enabled_sensors.append(sbs_id.Sensors.SPAR)
    if instrument_info["NmeaPositionDataAdded"] == "1":
        enabled_sensors.append(sbs_id.Sensors.nmeaLocation)
    if instrument_info["NmeaDepthDataAdded"] == "1":
        enabled_sensors.append(sbs_id.Sensors.nmeaDepth)
    if instrument_info["NmeaTimeAdded"] == "1":
        enabled_sensors.append(sbs_id.Sensors.nmeaTime)
    else:
        enabled_sensors.append(sbs_id.Sensors.SystemTime)

    # handle Scanfish data
    if not [line for line in hex.sbe9_data if line.startswith("SBE 11plus")]:
        enabled_sensors.append(sbs_id.Sensors.SystemTime)

    # use own function to read hex file
    with open(hex.path_to_file, "r") as f:
        result = f.readlines()

    array_data = []
    for line in result:
        if line.startswith("*"):
            continue
        hex_data = sbs_id.read_hex(
            instrument_type,
            line,
            enabled_sensors,
        )
        array_data.append(hex_data)

    data = pd.DataFrame(array_data)

    return data


def sorting_parameters(
    sensor_info: list,
    rule: list = [
        "Pressure",
        "Temperature",
        "Temperature2",
        "Conductivity",
        "Conductivity2",
    ],
) -> list:
    out_list = []
    for name in rule:
        for param in sensor_info:
            if name == param["SensorName"]:
                out_list.append(param)

    for param in sensor_info:
        if param["SensorName"] not in rule:
            out_list.append(param)

    return out_list


def get_time_gaps(raw_data: pd.DataFrame):
    data_integrity = raw_data["data integrity"].values.astype(int)
    diff = np.diff(data_integrity) % 256
    gaps = np.where(diff != 1)[0]
    gap_sizes = {a: diff[a] - 1 for a in gaps}
    return gap_sizes


def handle_time(
    gap_sizes: dict,
    hex: HexFile,
    data_size: int,
    parameters: Parameters = Parameters([], [], True),
):
    for index, gap_size in sorted(gap_sizes.items(), reverse=True):
        if index > data_size:
            continue
        for param in parameters.values():
            param.data = np.insert(param.data, index, [np.nan] * (gap_size))

    seconds_since_start = np.cumsum(
        np.concatenate(
            [
                [0.0],
                np.ones((data_size + sum(gap_sizes.values())) - 1) * (1 / 24),
            ]
        )
    )

    parameters.create_parameter(
        np.round(seconds_since_start, 3),
        {
            "shortname": "timeS",
            "longinfo": "Time, Elapsed [seconds]",
            "name": "Time",
            "unit": "seconds",
            "metainfo": "Time",
        },
    )

    start_time_posix = hex.start_time.timestamp()
    corrected_time_array = seconds_since_start + start_time_posix

    parameters.create_parameter(
        corrected_time_array.astype("int"),
        {
            "shortname": "timeU",
            "longinfo": "Unix Timestamp [s]",
            "name": "Unix Timestamp",
            "unit": "s",
            "metainfo": "Unix Timestamp",
        },
    )


def decode_hex(
    hex: HexFile | Path | str,
    xmlcon: XMLCONFile | Path | str = "",
    downcast_only: bool = True,
    **kwargs,
) -> CTDData:
    # input check
    if not isinstance(hex, HexFile):
        try:
            hex = HexFile(hex)
        except Exception as error:
            message = f"Could not open hex file {hex}: {error}"
            logger.error(message)
            sys.exit(message)

    if xmlcon:
        if not isinstance(xmlcon, XMLCONFile):
            try:
                xmlcon = XMLCONFile(xmlcon)
            except Exception as error:
                message = f"Could not open xmlcon file {xmlcon}: {error}"
                logger.warning(message)
        hex.xmlcon = xmlcon

    if not hex.xmlcon:
        sys.exit(
            f"No corresponding xmlcon for hex file {hex} found. Aborting."
        )
    parameters = Parameters([], [], True)
    cast_borders = {}
    raw_data = hex_reading(hex)
    for sensor in sorting_parameters(hex.xmlcon.sensor_info):
        sensor_name = sensor["SensorName"]
        if sensor_name in ["FluoroWetlabCDOM"]:
            continue
        mapping = ParameterMapping(sensor, raw_data, parameters)
        try:
            parameters.create_parameter(
                data=mapping.converted_data,
                metadata=mapping.metadata,
                name=sensor_name,
            )
        except AttributeError:
            logger.debug(f"{sensor_name} had no succesfull mapping.")

    # add lat and lon column
    if "NMEA Latitude" in raw_data.columns:
        parameters.create_parameter(
            raw_data["NMEA Latitude"],
            {
                "shortname": "latitude",
                "longinfo": "Latitude [deg]",
                "name": "Latitude",
                "unit": "deg",
                "metainfo": "latitude",
            },
        )
        parameters.create_parameter(
            raw_data["NMEA Longitude"],
            {
                "shortname": "longitude",
                "longinfo": "Longitude [deg]",
                "name": "Longitude",
                "unit": "deg",
                "metainfo": "longitude",
            },
        )

    # add flag column
    data_length = raw_data.shape[0]
    parameters.create_parameter(data=np.zeros(data_length), name="flag")
    # correct time
    gap_sizes = get_time_gaps(raw_data)
    handle_time(gap_sizes, hex, data_length, parameters)
    # cast borders
    if downcast_only:
        cast_borders = get_cast_borders(
            parameters["prDM"].data,
            downcast_only,
            **kwargs,
        )
        cast_borders["input_size"] = raw_data.shape[0]
        for param in parameters.values():
            param.data = param.data[
                cast_borders["down_start"] : cast_borders["down_end"] + 1
            ]
    # create processing_steps
    timestamp = datetime.now(timezone.utc).strftime("%Y.%m.%d %H:%M:%S")
    try:
        version = f", v{importlib.metadata.version('ctdam.parser')}"
    except Exception:
        version = ""
    processing_steps = CnvProcessingSteps([])
    processing_steps.add_info(
        module="hex2py",
        key="metainfo",
        value=f"{timestamp}, ctdam.parser python package{version}",
    )
    if gap_sizes:
        processing_steps.add_info(
            module="hex2py",
            key="time_correction",
            value=", ".join(
                [
                    f"{str(key)}: {str(value)}"
                    for key, value in gap_sizes.items()
                ]
            ),
        )
    if cast_borders:
        processing_steps.add_info(
            module="hex2py",
            key="cast_borders",
            value=", ".join(
                [
                    f"{str(key)}: {str(value)}"
                    for key, value in cast_borders.items()
                    if key.strip().startswith(("down", "up", "input_size"))
                ]
            ),
        )

    # CTDData instance to collect all info
    return CTDData(
        parameters=parameters,
        metadata_source=hex,
        processing_steps=processing_steps,
    )



import logging
import warnings
from typing import Tuple

import numpy as np
from scipy.signal import find_peaks, find_peaks_cwt, savgol_filter

logger = logging.getLogger(__name__)


def smoothing(x: np.ndarray) -> np.ndarray:
    size = x.shape[0]
    window_length = size // 40
    if window_length < 500:
        window_length = 500
    elif window_length > 2000:
        window_length = 2000
    for window in [
        window_length,
        window_length // 2,
        window_length // 10,
        size // 40,
        size // 100,
        25,
    ]:
        try:
            return savgol_filter(x, window_length=window, polyorder=3)
        except ValueError:
            continue
    return x


def get_cast_borders(
    pressure: np.ndarray,
    downcast_only: bool = True,
    min_size_factor: float = 0.01,
    min_soak_window: int = 100,
    max_fd_quotient: int = 6,
    prominence_divisor: int = 7,
    win_size_divisor: int = 500,
    min_velocity_quotient: int = 15,
    min_velocity: float = 0.045,
) -> dict:
    """Gets the borders of a given cast."""

    out_dict = {}

    # calculate first and second derivative
    smoothed_pressure = smoothing(pressure)
    maximum_pressure_index = np.nanargmax(smoothed_pressure)
    pressure_to_max = smoothed_pressure[:maximum_pressure_index]

    try:
        first_derivative = smoothing(np.gradient(pressure_to_max) * 24)
    except ValueError:
        out_dict["down_start"] = 0
        out_dict["down_end"] = maximum_pressure_index
        return out_dict

    second_derivative = smoothing(np.gradient(first_derivative))

    down_start, out_dict["fd_minima"], out_dict["sd_maxima"] = (
        get_downcast_start(
            first_derivative,
            second_derivative,
            base_data_size=pressure.shape[0],
            min_soak_window=min_soak_window,
            max_fd_quotient=max_fd_quotient,
            prominence_divisor=prominence_divisor,
            win_size_divisor=win_size_divisor,
            min_velocity_quotient=min_velocity_quotient,
            min_velocity=min_velocity,
        )
    )
    down_end = get_downcast_end(
        smoothed_pressure, first_derivative, second_derivative
    )

    # last sanity check
    with warnings.catch_warnings(action="ignore"):
        if (
            np.mean(first_derivative[down_start - 480 : down_start])
            > min_velocity * 2
        ):
            down_start = (
                out_dict["sd_maxima"][0]
                if len(out_dict["sd_maxima"]) > 0
                else 0
            )

    if (
        down_end - down_start < pressure.shape[0] * min_size_factor
        or down_end - down_start < 200
    ):
        warnings.warn(
            f"Found cast borders below the minimum cast size threshold of {pressure.shape[0] * min_size_factor}, defaulting to full cast size.",
            RuntimeWarning,
        )
        down_start = 0
        down_end = maximum_pressure_index

    out_dict["down_start"] = down_start
    out_dict["down_end"] = down_end

    if not downcast_only:
        out_dict["up_start"] = get_upcast_start(
            out_dict["down_end"], smoothed_pressure
        )
        out_dict["up_end"] = get_upcast_end(
            out_dict["down_end"], smoothed_pressure
        )

    return out_dict


def get_downcast_end(
    smoothed_pressure: np.ndarray,
    first_derivative: np.ndarray,
    second_derivative: np.ndarray,
) -> int:
    """
    Gets the downcast end of a given cast, accounting for heave due to waves.
    Returns the index of the highest pressure where the descent rate is below min_descent_rate.

    Parameters:
    -----------
    pressure : np.ndarray
        The pressure array.

    Returns:
    --------
    The index of the end of the downcast.
    """
    maximum_pressure_index = np.nanargmax(smoothed_pressure)
    pressure_border = [
        index
        for index, value in enumerate(smoothed_pressure)
        if value > (smoothed_pressure[maximum_pressure_index] - 2)
    ]

    if pressure_border:
        lower_pressure_border = pressure_border[0]
    else:
        min_fd_index = np.nanargmin(first_derivative)
        max_sd_index = np.nanargmax(second_derivative[min_fd_index:])
        lower_pressure_border = max_sd_index + min_fd_index

    min_sd_index = np.nanargmin(second_derivative[lower_pressure_border:])
    return int(min_sd_index + lower_pressure_border)


def get_downcast_start(
    first_derivative: np.ndarray,
    second_derivative: np.ndarray,
    base_data_size: int,
    min_soak_window: int = 100,
    max_fd_quotient: int = 6,
    prominence_divisor: int = 7,
    win_size_divisor: int = 500,
    min_velocity_quotient: int = 15,
    min_velocity: float = 0.045,
) -> Tuple[int, list, list]:
    """
    Gets the downcast start of a given cast, removing soaking/waiting time.
    Returns the index where the CTD begins to continuously move downward.

    Parameters:
    -----------
    pressure : np.ndarray
        The pressure array.

    Returns:
    --------
    The index of the start of the downcast.
    """
    max_fd = np.nanmax(first_derivative)

    # detect fluctuations in decent rate
    # if none found, set downcast start point to 0
    prominent_minimum = 0
    all_fd_minima = list(reversed(find_peaks_cwt(-first_derivative, 24)))
    all_maxima = []
    window_half = min_soak_window // 2
    for minimum in all_fd_minima:
        with warnings.catch_warnings(action="ignore"):
            if (
                np.mean(
                    first_derivative[
                        minimum - window_half : minimum + window_half
                    ]
                )
                < -max_fd / max_fd_quotient
            ):
                all_maxima.append(minimum)
                prominent_minimum = minimum

    down_start = 0
    all_sd_maxima = []

    if prominent_minimum:
        first_derivative = first_derivative[prominent_minimum:]
        all_sd_maxima = find_peaks(
            second_derivative[prominent_minimum:],
            prominence=np.nanmax(second_derivative[prominent_minimum:])
            / prominence_divisor,
        )[0]

        search_window = base_data_size // win_size_divisor

        for maximum in sorted(all_sd_maxima):
            mean = np.nanmean(
                first_derivative[maximum : maximum + search_window]
            )
            if mean > max_fd / min_velocity_quotient:
                down_start = maximum + prominent_minimum
                break
            for i in range(search_window * 7):
                index = maximum + i
                if index < len(first_derivative):
                    if first_derivative[index] > 0:
                        mean = np.nanmean(
                            first_derivative[index : index + search_window]
                        )
                        if mean > min_velocity:
                            down_start = index + prominent_minimum
                            break
            if down_start:
                break

    return (
        int(down_start),
        all_maxima,
        [maxi + prominent_minimum for maxi in all_sd_maxima],
    )


def get_upcast_start(ind_dc_end: int, smooth_velo: np.ndarray) -> int | None:
    upcast_velo_mean = np.mean(smooth_velo[ind_dc_end : len(smooth_velo)])
    for i in range(ind_dc_end, len(smooth_velo)):
        if smooth_velo[i] < upcast_velo_mean * 0.5:
            return i
    logger.warning("Could not find the upcast start.")
    return None


def get_upcast_end(ind_dc_end: int, smooth_velo: np.ndarray) -> int | None:
    upcast_velo_mean = np.mean(smooth_velo[ind_dc_end : len(smooth_velo)])
    for i in range(len(smooth_velo) - 1, ind_dc_end, -1):
        if smooth_velo[i] < upcast_velo_mean * 0.5:
            return i
    logger.warning("Could not find the upcast end.")
    return None



import importlib.metadata
import logging
from abc import ABC, abstractmethod
from datetime import datetime, timezone
from pathlib import Path

import numpy as np
import pandas as pd

from ctdam.parser import CnvFile
from ctdam.parser.ctddata import CTDData
from ctdam.proc.utils import BinnedDataError

logger = logging.getLogger(__name__)


class Module(ABC):
    """
    An interface to implement new processing modules against.

    Is meant to perform and unify all the necessary work that is needed to
    streamline processing on Sea-Bird CTD data. Implementing classes should
    only overwrite the transformation method, that does the actual altering of
    the data. All other organizational overhead should be covered by this
    interface. This includes parsing to .cnv output with correct handling of
    the metadata header.
    """

    def __init__(self) -> None:
        self.parent_module = "ctdam"
        self.info = self.__doc__
        self.name = self.__class__.__name__.lower()

    def __call__(
        self,
        arguments: dict,
        default_values: dict = {},
        output_name: str | None = None,
    ) -> None | CnvFile | CTDData | pd.DataFrame | np.ndarray:
        self.arguments = {**default_values, **arguments}
        self.output_name = output_name

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return self.__str__()

    def add_processing_metadata(self):
        """
        Parses the module processing information into cnv-compliant metadata
        lines.

        These take on the form of {MODULE_NAME}_{KEY} = {VALUE} for every
        key-value pair inside of the given dictionary with the modules
        processing info.

        """
        # if isinstance(self.cnv, CnvFile):
        if hasattr(self, "processing_steps"):
            # general header for every module
            timestamp = datetime.now(timezone.utc).strftime(
                "%Y.%m.%d %H:%M:%S"
            )
            try:
                version = (
                    f", v{importlib.metadata.version(self.parent_module)}"
                )
            except Exception:
                version = ""
            self.processing_steps.add_info(
                module=self.name,
                key="metainfo",
                value=f"{timestamp}, {self.parent_module} python package{version}",
            )
            for key, value in self.arguments.items():
                if key == "file_suffix":
                    continue
                self.processing_steps.add_info(
                    module=self.name,
                    key=key,
                    value=value,
                )
        else:
            logger.error(
                "Cannot write processing metainfo without any cnv source."
            )

    def load_file(self, file_path: Path) -> CnvFile:
        """
        Loads the target files information into an CnvFile instance.

        Parameters
        ----------
        file_path: Path :
            Path to the target file.

        Returns
        -------
        CnvFile object representing the file in the file system.

        """
        return CnvFile(file_path)

    @abstractmethod
    def to_cnv(self):
        pass


class ArrayModule(Module):
    def __init__(self) -> None:
        super().__init__()

    def __call__(
        self,
        input: Path | str | CnvFile | CTDData,
        arguments: dict = {},
        output: str = "cnvobject",
        output_name: str | None = None,
        default_values: dict = {},
        original_input_path: Path | str | None = None,
        bad_flag: float = -9.990e-29,
    ) -> None | CnvFile | CTDData:
        super().__call__(arguments, default_values, output_name)
        if original_input_path:
            self.original_input_path = Path(original_input_path)
        if isinstance(input, Path | str):
            self.ctd_data = CnvFile(
                input, create_dataframe=False
            ).to_ctd_data()
            self.original_input_path = Path(input)
        elif isinstance(input, CnvFile):
            self.ctd_data = input.to_ctd_data()
            self.original_input_path = input.path_to_file
        elif isinstance(input, CTDData):
            self.ctd_data = input
        else:
            raise TypeError(f"Incorrect input type: {type(input)}. Aborting.")
        self.processing_steps = self.ctd_data.processing_steps
        self.sample_rate = self.ctd_data.sample_rate
        self.bad_flag = bad_flag
        self.create_flag_array_if_missing()
        self.flags = self.ctd_data["flag"].data.astype(bool)
        self.ran_processing = self.transformation()
        try:
            self.ctd_data["flag"].data = self.flags.astype(float)
        except KeyError:
            pass
        if self.ran_processing:
            self.add_processing_metadata()
        if "file_suffix" in self.arguments:
            output = "cnv"
        if output.lower() in ("cnv", "file"):
            self.to_cnv()
        return self.ctd_data

    @abstractmethod
    def transformation(self) -> bool:
        pass

    def get_array(self) -> np.ndarray:
        return self.ctd_data.get_full_data_array()

    def get_data_size(self) -> int:
        return self.get_array().shape[0]

    def create_flag_array_if_missing(self):
        if self._check_parameter_existence("flag"):
            return
        self.ctd_data.create_parameter(
            data=np.zeros(self.get_data_size()), name="flag"
        )

    def handle_new_flags(self, new_flag_array: np.ndarray):
        # print out the number of flagged data rows
        new_flag_count = new_flag_array.sum()
        row_count = self.get_data_size()
        self.arguments["bad_rows"] = f"{new_flag_count} of {row_count} ({
            new_flag_count / row_count:1.2f})"
        # incorporate the new flags
        self.flags |= new_flag_array

    def check_whether_working_on_binned_data(self):
        if self.ctd_data.binned:
            raise BinnedDataError(
                file_name=self.ctd_data.file_name,
                step_name=self.name,
            )

    def _check_parameter_existence(
        self, parameter: str, shortname: bool = True
    ) -> bool:
        """
        Helper method to ensure parameter presence in input data before
        attempting the transformation.

        Parameters
        ----------
        parameter: str :
            The parameter to check for.

        Returns
        -------
        Whether the parameter is present inside of the cnv parameters or not.

        """
        if shortname:
            return parameter in self.ctd_data.parameters
        else:
            return parameter in [p.param for p in self.ctd_data]

    def to_cnv(self):
        if not self.ctd_data:
            return
        self.ctd_data.parameters.full_data_array = self.array
        if "file_suffix" in self.arguments:
            if self.output_name:
                output_name = Path(self.output_name)
                stem = output_name.stem
                self.output_name = output_name.with_stem(
                    stem + self.arguments["file_suffix"]
                )
            else:
                self.output_name = self.ctd_data.path_to_file.with_stem(
                    self.ctd_data.file_name + self.arguments["file_suffix"]
                )
        self.ctd_data.to_cnv(self.output_name)


class DataFrameModule(Module):
    def __call__(
        self,
        input: Path | str | CnvFile | pd.DataFrame | np.ndarray,
        arguments: dict = {},
        output: str = "cnvobject",
        output_name: str | None = None,
    ) -> None | CnvFile | pd.DataFrame | np.ndarray:
        super().__call__(arguments, output_name)
        if isinstance(input, Path | str):
            self.cnv = self.load_file(Path(input))
            self.df = self.cnv.df
        elif isinstance(input, CnvFile):
            self.cnv = input
            self.df = self.cnv.df
        elif isinstance(input, pd.DataFrame):
            self.cnv = None
            self.df = input
        else:
            raise TypeError(f"Incorrect input type: {type(input)}. Aborting.")
        self.df = self.transformation()
        self.add_processing_metadata()
        if output.lower() in ("cnv", "file"):
            self.to_cnv()
            return None
        elif output.lower() in ("internal", "cnvobject") and isinstance(
            self.cnv, CnvFile
        ):
            return self.cnv
        else:
            return self.df

    @abstractmethod
    def transformation(self) -> pd.DataFrame:
        """
        The actual data transformation on the CTD data.

        Needs to be implemented by the implementing classes.
        """
        df = self.df
        return df

    def _check_parameter_existence(self, parameter: str) -> bool:
        """
        Helper method to ensure parameter presence in input data before
        attempting the transformation.

        Parameters
        ----------
        parameter: str :
            The parameter to check for.

        Returns
        -------
        Whether the parameter is present inside of the cnv dataframe or not.

        """
        # ensure shortnames as column names
        self.df.meta.header_detail = "shortname"
        return parameter in self.df.columns

    def to_cnv(
        self,
        additional_data_columns: list[str] = [],
        custom_data_columns: list | None = None,
    ):
        """
        Writes the internal CnvFile instance to disk.

        Uses the CnvFile's output parser for that and organizes the different
        bits of information for that.

        Parameters
        ----------
        additional_data_columns: list[str] :
            A list of columns that in addition to the ones inside the original
            dataframe.
             (Default value = [])
        custom_data_columns: list | None :
            A list of coulumns that will exclusively used to select the data
            items for the output .cnv .
             (Default value = None)

        """
        if isinstance(self.cnv, CnvFile):
            if custom_data_columns:
                header_list = custom_data_columns
            else:
                header_list = [
                    header[self.cnv.df.meta.header_detail]
                    for header in list(self.cnv.df.meta.metadata.values())
                ]
            self.cnv.df = self.df
            self.cnv.to_cnv(
                file_name=self.output_name,
                header_list=[*header_list, *additional_data_columns],
            )
        else:
            logger.error("Cannot write to cnv without any cnv as source.")

    def to_csv(self):
        """Writes the dataframe as .csv to disk."""
        try:
            self.df.to_csv()
        except IOError as error:
            logger.error(f"Failed to write dataframe to csv: {error}")


class MissingParameterError(Exception):
    """A custom error to throw when necessary parameters are missing from the
    input .cnv file."""

    def __init__(self, step_name: str, parameter_name: str):
        super().__init__(
            f"Could not run processing step {
                step_name
            } due to a missing parameter: {parameter_name}"
        )



import tkinter as tk
import tkinter.font as tkFont
from inspect import isfunction
from pathlib import Path
from tkinter import ttk
from typing import Callable

try:
    import customtkinter as ctk
    from CTkMessagebox import CTkMessagebox
except ImportError:
    raise ImportError(
        "The 'gui' extra is required to use this feature. Install with: pip install ctd-ctdam.proc[gui]"
    )

from ctdam.entry.gui import TomlEditor
from ctdam.proc import Module
from ctdam.proc.modules.available_modules import (
    get_dict_of_available_processing_modules,
)
from ctdam.proc.modules.external_functions import ExternalFunctionInfo


class ModuleSelector(ctk.CTkToplevel):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.grid_columnconfigure(0, weight=1)
        self.grid_rowconfigure(2, weight=1)
        self.padx = 5
        self.pady = 5
        self.title("Select a processing module from the list")
        self.heading = ctk.CTkLabel(
            self,
            text="Available Modules",
            font=(tkFont.nametofont("TkDefaultFont"), 35),
        )
        self.heading.grid(row=0, column=0, padx=15, pady=15, sticky="N")
        self.content_frame = ctk.CTkScrollableFrame(
            self, fg_color="transparent"
        )
        self.content_frame.grid(row=2, column=0, sticky="nsew")
        self.content_frame.grid_columnconfigure(0, weight=1)
        self.content_frame.grid_rowconfigure(0, weight=1)
        self.populate()

    def populate(self):
        for index, (category, modules) in enumerate(
            get_dict_of_available_processing_modules().items()
        ):
            category_frame = ctk.CTkFrame(
                self.content_frame, fg_color="transparent"
            )
            category_frame.grid(row=index, column=0)
            category_title = ctk.CTkLabel(
                category_frame,
                text=category,
                font=(tkFont.nametofont("TkDefaultFont"), 20),
            )
            category_title.grid(row=0, column=0)
            ttk.Separator(category_frame, orient="horizontal").grid(
                row=1, column=0, columnspan=3, sticky="ew"
            )
            if isinstance(modules, dict):
                for module_index, (module, module_info) in enumerate(
                    modules.items()
                ):
                    self.create_module_field(
                        category_frame, module, module_index + 2, module_info
                    )
            else:
                for module_index, module in enumerate(modules):
                    self.create_module_field(
                        category_frame, module, module_index + 2
                    )

    def create_module_field(
        self,
        frame,
        module,
        row,
        ex_fun_info: ExternalFunctionInfo | None = None,
    ):
        if isfunction(module):
            ex_fun_info = module
            module = module.__name__
        module_name = ctk.CTkLabel(
            frame,
            text=str(module),
        )
        module_name.grid(
            row=row,
            column=0,
            padx=self.padx,
            pady=self.pady,
        )
        if ex_fun_info:
            info = ex_fun_info
        else:
            info = module
        module_info = ctk.CTkButton(
            frame,
            text="Info",
            command=lambda: show_info(info),
        )
        module_info.grid(
            row=row,
            column=1,
            padx=self.padx,
            pady=self.pady,
        )
        module_select = ctk.CTkButton(
            frame,
            text="Select",
            command=lambda: select(str(module)),
        )
        module_select.grid(
            row=row,
            column=2,
            padx=self.padx,
            pady=self.pady,
        )

        def show_info(module: str | Module | ExternalFunctionInfo | Callable):
            if isinstance(module, Module):
                message = str(module.info).replace("\n", " ")
            elif isinstance(module, ExternalFunctionInfo):
                message = module.general_info
            elif isinstance(module, Callable):
                message = module.__doc__
                module = module.__name__
            else:
                message = (
                    f"No information for the module '{module}' available."
                )
            CTkMessagebox(
                title=f"Module: {str(module)}",
                message=message,
                option_1="Ok",
            )

        def select(module: str):
            self.master.add_module(module)
            for frame in self.winfo_children():
                frame.destroy()
            self.destroy()


class ProcedureConfigView(TomlEditor):
    """
    A frame that contains the configuration for a procedure.
    """

    def __init__(
        self,
        master: ctk.CTkFrame,
        title: str = "Processing Procedure Config Editor",
        possible_parameters: list[str] = [
            "psa_directory",
            "output_dir",
            "output_name",
        ],
        config_file: Path | str = "",
        title_size: int = 20,
        default_dir_to_save_in: Path | str = "",
    ):
        super().__init__(
            master,
            title,
            possible_parameters,
            config_file,
            title_size,
            default_dir_to_save_in=default_dir_to_save_in,
        )
        self.module_frames = {}
        self.new_module_frame = ctk.CTkFrame(self, fg_color="transparent")
        self.new_module_frame.grid(row=3, column=0)
        self.add_module_button = ctk.CTkButton(
            self.new_module_frame, text="Add Module", command=self.new_module
        )
        self.add_module_button.grid(row=0, column=1, padx=5, pady=10)

    def new_module(self):
        selector = ModuleSelector(master=self)
        selector.geometry("700x850")
        selector.grab_set()

    def remove_module(self, module_name, frame):
        frame.destroy()
        self.config_data["modules"].pop(module_name, None)
        self.module_frames.pop(module_name, None)

    def remove_module_param(
        self, module_name, frame, key_entry, value_entry, param_entries
    ):
        self.update_module_params(module_name, param_entries)
        parent_frame = frame.master
        key_entry.destroy()
        value_entry.destroy()
        frame.grid_forget()
        frame.destroy()
        parent_frame.grid()
        param_entries.pop(key_entry, None)

    def load_config_specific_data(self, row=0):
        if "modules" in self.config_data:
            for module_name, params in self.config_data["modules"].items():
                self.add_module(module_name, params, row=row)
                row += 1

    def add_module(self, module_name=None, params=None, row=None):
        frame = ctk.CTkFrame(self.content_frame, fg_color="transparent")
        frame.grid(row=row, column=0, sticky="ew", padx=5, pady=5)
        frame.grid_columnconfigure(0, weight=1)
        ttk.Separator(frame, orient="horizontal").grid(
            row=0, column=0, columnspan=3, sticky="ew"
        )
        header = ctk.CTkLabel(
            frame,
            text=f"Module: {module_name or 'New Module'}",
            font=("Arial", 14, "bold"),
        )
        header.grid(row=1, column=0, columnspan=2, sticky="w", padx=5, pady=5)

        if module_name is None:
            module_name = f"module_{len(self.module_frames) + 1}"

        self.module_frames[module_name] = frame

        param_entries = {}
        if params:
            for index, (key, value) in enumerate(params.items()):
                self.create_module_param_field(
                    module_name, frame, key, value, param_entries, index
                )
        add_param_button = ctk.CTkButton(
            frame,
            text="+",
            command=lambda: self.create_module_param_field(
                module_name, frame, "", "", param_entries
            ),
        )
        add_param_button.grid(row=1, column=2, sticky="e", padx=5, pady=5)

        remove_module_button = ctk.CTkButton(
            frame,
            text="Remove Module",
            command=lambda: self.remove_module(module_name, frame),
        )
        remove_module_button.grid(row=1, column=1, sticky="e", padx=5, pady=5)
        self.config_data.setdefault("modules", {})[module_name] = {}
        self.update_module_params(module_name, param_entries)
        frame.bind(
            "<FocusOut>",
            lambda e: self.update_module_params(module_name, param_entries),
        )

    def update_module_params(self, module_name, param_entries):
        self.config_data["modules"][module_name] = {
            k.get(): v.get() for k, v in param_entries.items()
        }

    def create_module_param_field(
        self,
        module_name: str,
        parent_frame: ctk.CTkFrame,
        key: str,
        value: str,
        param_entries: dict,
        index=None,
    ):
        if not index:
            index = len(parent_frame.winfo_children())
        frame = ctk.CTkFrame(parent_frame, fg_color="transparent")

        def reload_frame(e):
            frame.destroy()
            try:
                param_entries.pop(key_entry)
            except KeyError:
                pass
            self.create_module_param_field(
                module_name, parent_frame, e, value, param_entries, index
            )

        key_entry = ctk.CTkComboBox(
            frame,
            values=["psa", "file_suffix", "bl", "Oxygen"],
            command=reload_frame,
        )

        key_entry.grid(row=0, column=0, padx=20, pady=5)
        key_entry.set(key)

        value_entry = ctk.CTkEntry(frame)

        def update_param(event=None):
            param_entries[key_entry] = value_entry
            self.update_module_params(module_name, param_entries)

        if key in ["psa", "bl"]:
            file_picker = self.create_picker_element(
                frame=frame,
                entry=value_entry,
                directory=False,
                callback=update_param,
            )
            file_picker.grid(row=0, column=1)
            width = 340
        else:
            width = 400

        value_entry.grid(row=0, column=2, padx=5, pady=5)
        value_entry.configure(require_redraw=True, width=width)
        value_entry.insert(0, value)
        value_entry.xview(tk.END)

        remove_button = ctk.CTkButton(
            frame,
            text="-",
            command=lambda: self.remove_module_param(
                module_name, frame, key_entry, value_entry, param_entries
            ),
        )
        remove_button.grid(row=0, column=3, padx=5, pady=5)
        param_entries[key_entry] = value_entry

        key_entry.bind("<FocusOut>", update_param)
        value_entry.bind("<FocusOut>", update_param)
        frame.grid(row=index + 1, column=0, columnspan=3, sticky="ew")

    def check_input(self) -> bool:
        # TODO: think about sophisticated check here
        return True


def run_gui(file_to_open: str = "proc_test.toml"):
    # Example usage
    ctk.set_appearance_mode("System")
    ctk.set_default_color_theme("blue")

    app = ctk.CTk()
    app.geometry("800x600")

    editor = ProcedureConfigView(
        app,
        config_file=file_to_open,
        title_size=35,
    )

    editor.grid(row=0, column=0, sticky="nsew")

    # Configure the grid to make the editor expand with the window
    app.grid_rowconfigure(0, weight=1)
    app.grid_columnconfigure(0, weight=1)
    app.mainloop()


if __name__ == "__main__":
    run_gui()



import logging
from collections import UserDict
from inspect import getmembers, isfunction
from pathlib import Path
from typing import Callable

import docstring_parser
import numpy as np
import pandas as pd

from ctdam.parser.ctddata import CTDData
from ctdam.proc.module import ArrayModule

logger = logging.getLogger(__name__)


class ExternalFunctions(UserDict):
    def __init__(self, modules: list) -> None:
        self.data = {}
        for module in modules:
            self.data[module.__name__] = self.get_module_functions(module)

    def available_modules(self) -> list:
        return list(self.data.keys())

    def functions_of_certain_module(self, module: str) -> dict:
        if module in self.available_modules():
            return self.data[module]
        else:
            return {}

    def get_all_functions(self) -> dict:
        out_dict = {}
        for module in self.available_modules():
            out_dict = {**out_dict, **self.functions_of_certain_module(module)}
        return out_dict

    def list_of_function_names(self, module: str = "") -> list:
        if module:
            return [key for key in self.functions_of_certain_module(module)]
        else:
            return [key for key in self.get_all_functions()]

    def get_module_functions(self, module) -> dict:
        out_dict = {}
        for name, function in getmembers(module, isfunction):
            out_dict[name] = ExternalFunctionInfo(function)
        return out_dict


class ExternalFunctionInfo:
    def __init__(self, external_function: Callable) -> None:
        self.function = external_function
        self.name = self.function.__name__
        module = self.function.__module__
        self.module = module.split(".")[0] if "." in module else module
        self.raw_docstring = self.function.__doc__
        self.parse_docstring(self.raw_docstring)

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return self.__str__()

    def run(self, ctd_data: CTDData, parameters: dict = {}) -> bool:
        if parameters:
            return self._run_with_parameters(ctd_data, parameters)
        else:
            return self._run_with_mapping(ctd_data)

    def _run_with_parameters(
        self,
        ctd_data: CTDData,
        parameters: dict,
    ) -> bool:
        try:
            self.execute_funtion(
                args=list(parameters.values()), ctd_data=ctd_data
            )
        except Exception as error:
            logger.warning(f"Could not run {self.name}: {error}")
            return False
        return True

    def _run_with_mapping(self, ctd_data: CTDData) -> bool:
        args0 = []
        args1 = []
        second_column = False
        for parameter in self.param_info:
            try:
                columns = self.map_parameter(parameter["name"], ctd_data)
                if len(columns) == 0:
                    logger.warning(
                        f"Could not run {self.name}, argument {parameter} was not understood."
                    )
                    return False
                elif len(columns) == 1:
                    args0.append(ctd_data.parameters[columns[0]].data)
                    args1.append(ctd_data.parameters[columns[0]].data)
                elif len(columns) == 2:
                    args0.append(ctd_data.parameters[columns[0]].data)
                    args1.append(ctd_data.parameters[columns[1]].data)
                    second_column = True
                else:
                    raise ValueError(
                        f"Unexpected number of columns in: {columns}"
                    )
            except KeyError as error:
                logger.warning(
                    f"Could not run {self.name} without column {str(error).strip()}. "
                )
                return False

        return_value0 = self.execute_funtion(args0, ctd_data)
        if second_column:
            return_value1 = self.execute_funtion(args1, ctd_data, True)
        else:
            return_value1 = True
        return return_value0 and return_value1

    def execute_funtion(
        self,
        args: list,
        ctd_data: CTDData,
        second_sensor: bool = False,
    ) -> bool:
        try:
            new_columns = self.function(*args)
        except Exception as error:
            logger.warning(f"Could not run {self.name}: {error}")
            return False
        else:
            if not len(new_columns.shape) == len(self.return_info):
                logger.warning(
                    f"Could not run {self.name}: output was not expected."
                )
                return False
            for column, return_value in zip(new_columns, self.return_info):
                metadata = self.create_cnv_metadata(
                    return_value, second_sensor
                )
                ctd_data.parameters.create_parameter(
                    data=column,
                    metadata=metadata,
                    name=metadata["name"],
                )
        return True

    def create_cnv_metadata(
        self,
        return_value: dict,
        second_sensor: bool = False,
    ) -> dict:
        return_name = str(return_value["name"])
        mapped_name = self.map_parameter(return_name)
        if len(mapped_name) > 1:
            shortname = mapped_name[int(second_sensor)]
        else:
            shortname = f"{self.module}_{mapped_name[0].split('_')[0] if '_' in mapped_name[0] else mapped_name[0]}_{int(second_sensor)}"
        name = return_name.strip()
        unit = (
            return_value["type"].split(",")[1]
            if "," in return_value["type"]
            else return_value["type"]
        ).strip()
        metainfo = return_value["desc"].strip()
        return {
            "shortname": shortname,
            "longinfo": f"{name}, {metainfo} [{unit}]",
            "name": name,
            "metainfo": metainfo,
            "unit": unit,
        }

    def map_parameter(
        self,
        parameter: str,
        ctd_data: CTDData | None = None,
    ) -> list:
        mapper = {
            "p": ["prDM"],
            "SA": ["gsw_saA0", "gsw_saA1"],
            "SA_baltic": ["gsw_saA0", "gsw_saA1"],
            "CT": ["gsw_ctA0", "gsw_ctA1"],
            "t": ["t090C", "t190C"],
            "lat": ["latitude"],
            "lon": ["longitude"],
            "SP": ["sal00", "sal11"],
            "pt": ["potemp090C", "potemp190C"],
        }
        if parameter in mapper:
            return mapper[parameter]
        elif isinstance(ctd_data, CTDData):
            present_params = [
                p.name
                for p in ctd_data
                if p.param.lower() == parameter.lower()
            ]
            if present_params:
                return present_params
            else:
                return [parameter]
        else:
            return [parameter]

    def parse_docstring(self, raw_docstring):
        if not isinstance(raw_docstring, str):
            return None
        docstring = docstring_parser.parse(raw_docstring)
        if not docstring.style:
            return None
        self.general_info = str(docstring.short_description) + (
            str(docstring.long_description)
            if docstring.long_description
            else ""
        )
        self.param_info = [
            {
                "name": p.arg_name,
                "desc": p.description,
            }
            for p in docstring.params
        ]
        ret_object = docstring.returns
        if ret_object:
            self.return_info = [
                {
                    "name": ret_object.return_name,
                    "type": ret_object.type_name,
                    "desc": ret_object.description,
                }
            ]
        else:
            self.return_info = [{"name": self.name}]


class ExternalFunctionCaller(ArrayModule):
    """ """

    def __init__(
        self,
        module: str,
        processing_functions: ExternalFunctions,
    ) -> None:
        super().__init__()
        self.module = module
        if self.module not in processing_functions.list_of_function_names():
            raise ValueError(
                f"Could not run processing function: {module}, unkown."
            )
        self.function = processing_functions.get_all_functions()[module]
        self.name = self.function.name

    def __call__(
        self,
        input: Path | str | CTDData | pd.DataFrame | np.ndarray,
        arguments: dict = {},
        output: str = "cnvobject",
        output_name: str | None = None,
        **kwargs,
    ) -> None | CTDData | pd.DataFrame | np.ndarray:
        return super().__call__(input, arguments, output, output_name)

    def transformation(self) -> bool:
        self.parent_module = self.function.module
        try:
            return_value = self.function.run(self.ctd_data, self.arguments)
        except Exception as error:
            logger.warning(
                f"Could not run processing function: {self.module}: {error}"
            )
            return False
        return return_value



import logging
import multiprocessing
import sys
import warnings
from collections import UserList
from pathlib import Path

import pandas as pd
from tqdm import tqdm

from ctdam.conv import decode_hex
from ctdam.parser import CnvFile, CTDData
from ctdam.proc import Procedure
from ctdam.utils import get_unique_sensor_data
from ctdam.vis import basic_bokeh_plot, create_main_html

logger = logging.getLogger(__name__)


class Casts(UserList):
    """
    A structure to ease working with multiple ctd casts.

    Can work with ascii files, converted (.cnv) or non-converted (.hex), as
    well as data within python objects (CTDData). Automates the very basic
    actions ussually performs on these data: converting, processing, plotting
    and exporting. The cpu-heavy actions (convertion and processing) are
    calculated in parallel, using multithreading.
    """

    def __init__(
        self,
        path_to_data: Path | str = "",
        ctd_data: list[CTDData] = [],
        processing_info: dict = {},
        pattern: str = "",
        plot: bool = False,
        show_plot: bool = True,
        plot_dir: Path | str = "htmls",
    ):
        self.processing_info = processing_info
        self.plot_dir = plot_dir
        if ctd_data:
            self.data = ctd_data
            if [p.path_to_file.parent for p in ctd_data]:
                self.path_to_data = ctd_data[0].path_to_file.parent
            else:
                self.path_to_data = None
        elif path_to_data:
            self.path_to_data = Path(path_to_data)
            if self.path_to_data.is_dir():
                files = sorted(list(self.path_to_data.rglob(f"*{pattern}*")))
                cnvs = [f for f in files if f.suffix == ".cnv"]
                hexes = [f for f in files if f.suffix == ".hex"]
                if len(hexes) < len(cnvs):
                    self.data = [CnvFile(file).to_ctd_data() for file in cnvs]
                else:
                    with multiprocessing.Pool() as pool:
                        self.data = list(
                            tqdm(
                                pool.imap_unordered(self.convert, hexes),
                                total=len(hexes),
                                desc="Cast conversion",
                                unit="cast",
                            )
                        )
            elif self.path_to_data.is_file():
                if self.path_to_data.suffix == "cnv":
                    self.data = [CnvFile(self.path_to_data).to_ctd_data()]
                else:
                    self.data = [self.convert(self.path_to_data)]
            else:
                sys.exit(f"Invalid input path: {path_to_data}")
            self.anomalous_data = self.check_converted_data()
        else:
            sys.exit(f"Invalid input: {ctd_data} or {path_to_data}")
        self.cruise = (
            self.data[0].cruise if hasattr(self.data[0], "cruise") else ""
        )
        if processing_info:
            self.process(processing_info)
        if plot:
            self.plot(show_plot)
        self.data = sorted(self.data)

    def convert(self, file: Path):
        arguments = {}
        if "modules" in self.processing_info:
            if "hex2py" in self.processing_info["modules"]:
                arguments = self.processing_info["modules"]["hex2py"]
        try:
            with warnings.catch_warnings(action="ignore"):
                return decode_hex(file, **arguments)
        except Exception as error:
            logger.error(f"Could not convert file {file}: {error}")

    def check_converted_data(self) -> list[CTDData]:
        self.data = [c for c in self.data if c is not None]
        anomalies = []
        for cast in self.data:
            if cast.binned:
                if cast.get_data_length() < 10:
                    anomalies.append(cast)
            else:
                if cast.get_data_length() < 500:
                    anomalies.append(cast)
            try:
                if (
                    cast.cast_borders["down_end"]
                    / cast.cast_borders["input_size"]
                    < 0.01
                ) and cast not in anomalies:
                    anomalies.append(cast)
            except KeyError:
                continue
        self.data = [c for c in self.data if c not in anomalies]
        return anomalies

    def read_sensor_info(self):
        self.sensor_info = get_unique_sensor_data(
            [cast.sensor_info for cast in sorted(self.data)]
        )
        if len(self.sensor_info) > 1:
            warnings.warn(
                "Sensor anomalies found. Please check the sensor_info attribure for details."
            )
            sensor_info = []
            for configuration in self.sensor_info:
                for sensor in configuration[1]:
                    sensor_info.append(sensor)
        else:
            sensor_info = self.sensor_info[0][1]

    def process(self, processing_info: dict, target_files: list[CTDData] = []):
        target_files = target_files if target_files else self.data
        with multiprocessing.Pool() as pool:
            procedure = Procedure(processing_info, auto_run=False)
            self.data = list(
                tqdm(
                    pool.imap_unordered(procedure.run, target_files),
                    total=len(target_files),
                    desc="Processing",
                    unit="cast",
                )
            )

    def plot(self, show_plot: bool = True):
        html_directory = str(self.path_to_data.parent.joinpath(self.plot_dir))
        for cast in self.data:
            try:
                basic_bokeh_plot(
                    ctd_data=cast,
                    output_directory=html_directory,
                    print_plot=True,
                    metadata=True,
                    show_plot=show_plot,
                )
            except Exception:
                continue

        create_main_html(
            directory_path=html_directory,
            title=self.cruise,
        )

    def to_tsv(self, file_name: str | Path | None = None):
        file_name = f"{self.cruise}_CTD" if file_name is None else file_name

        if not hasattr(self, "df"):
            list_of_dfs = [cast.get_pandas_dataframe() for cast in self.data]
            self.df = pd.concat(list_of_dfs, ignore_index=True)
        df_out = self.df.copy()

        # apply number formatting for each parameter
        for column in df_out.columns:
            try:
                df_out[column] = df_out[column].astype("float")
            except ValueError:
                continue
            for param in self.data[0]:
                if column == param.name:
                    if param.name == "prDM":
                        formatter = "{:.1f}"
                    else:
                        formatter = param.output_format
                    df_out[column] = df_out[column].map(formatter.format)
        df_out.to_csv(
            path_or_buf=Path(file_name).with_suffix(".tab"),
            sep="\t",
            index=False,
        )



from pathlib import Path

import pandas as pd


class GEOMARCTDFile:
    """
    A parser to read .ctd files created by the GEOMAR ctdam.proc software.

    Goes through the file line by line and sorts the individual lines in
    corresponding lists. That way, data and different types of metadata are
    structured on a basic level.
    In general, this parser is meant to stick close to the way the Seabird-
    Parsers are written.
    """

    def __init__(
        self,
        path_to_file: Path | str,
        only_header: bool = False,
        create_dataframe: bool = True,
    ):
        self.path_to_file = Path(path_to_file)
        self.only_header = only_header
        self.raw_input = []
        self.metadata = {}
        self.history = []
        self.comment = []
        self.data_header = []
        self.raw_data = []
        self.read_file()
        if create_dataframe:
            self.create_dataframe()

    def __str__(self) -> str:
        return "/n".join(self.raw_data)

    def __repr__(self) -> str:
        return str(self.path_to_file.absolute())

    def __eq__(self, other) -> bool:
        return self.raw_data == other.raw_data

    def read_file(self):
        with open(self.path_to_file, "r") as file:
            past_header = False
            for line in file:
                self.raw_input.append(line)
                if line.startswith("History"):
                    self.history.append(
                        line.removeprefix("History  = # GEOMAR").strip()
                    )
                elif line.startswith("Comment"):
                    self.comment.append(
                        line.removeprefix("Comment  =").strip()
                    )
                elif line.startswith("Columns"):
                    self.data_header = [
                        column.removeprefix("Columns  =").strip()
                        for column in line.split(":")
                    ]
                    past_header = True
                    if self.only_header:
                        break
                else:
                    if not past_header:
                        try:
                            key, value = line.split("=")
                        except ValueError:
                            key = line
                            value = ""
                        self.metadata[key.strip()] = value.strip()
                    else:
                        self.raw_data.append(line)

    def create_dataframe(self):
        self.df = pd.DataFrame(
            [row.split() for row in self.raw_data],
            dtype=float,
            columns=self.data_header,
        )



import logging
import shutil
import sys
from datetime import datetime, timezone
from pathlib import Path

from ctdam.conv.hexdecoder import decode_hex
from ctdam.parser import CnvFile
from ctdam.parser.ctddata import CTDData
from ctdam.parser.hexfile import HexFile
from ctdam.proc.modules import map_proc_name_to_class
from ctdam.proc.seabird.module import ProcessingModule
from ctdam.proc.seabird.step import ProcessingStep
from ctdam.proc.settings import Configuration, IncompleteConfigFile
from ctdam.proc.utils import (
    default_seabird_exe_path,
    fill_file_type_dir,
)

try:
    from ctdam.vis.visualize import basic_bokeh_plot
except ImportError:
    plotting_available = False
else:
    plotting_available = True


logger = logging.getLogger(__name__)


class Procedure:
    """
    Runs a couple of processing steps in sequence on one or more CTD data
    source files.

    It can use seabird internal processing modules, as well as custom ones.
    These can be in the form of independent windows exes or just pure python
    code. The input data can be .hex, .cnv or pandas DataFrames. The input and
    all module and extra information is stored in a dict that usually will be
    generated by the settings Configuration module that reads a toml config.

    Parameters
    ----------
    configuration: dict | Configuration:
        The information necessary to run a processing procedure.

    seabird_exe_directory: Path | str | None = None:
        The path to the directory where the Sea-Bird exes reside in.
        Usually not necessary, as this Class knows the default install path.

    available_hex_converters: list[str] = ["datcnv"]:
        A list of the known hex converters.

    auto_run: bool = True:
        Whether to autopilot the whole procedure.

    procedure_fingerprint_directory: Path | str | None = None:
        A path to a directory where the fingerprint are meant to be stored in.
        If none given, this option is considered to be turned off.

    file_type_dir: Path | str | None = None:
        A path to a directory where the individual Sea-Bird file types are
        differentiated into respective directories.
        If none given, this option is considered to be turned off.

    verbose: bool = False:
        Sets whether the Sea-Bird modules are run silently or not.

    timeout: int = 60:
        The time in seconds after which individual processing steps will be
        killed automatically.

    Returns
    -------
    In auto_run mode, a .cnv file or an instance of CnvFile, depending on the
        file_type parameter inside of the configuration.
    Otherwise it is an invocation that collected and evaluated the information
        necessary to run a processing procedure on one or more target files.

    """

    def __init__(
        self,
        configuration: dict | Configuration,
        seabird_exe_directory: Path | str | None = None,
        available_hex_converters: list[str] = ["datcnv", "hex2py"],
        auto_run: bool = True,
        procedure_fingerprint_directory: Path | str | None = None,
        file_type_dir: Path | str | None = None,
        plot: bool = False,
        verbose: bool = False,
        timeout: int = 60,
    ) -> None:
        if isinstance(configuration, Configuration):
            self.config = configuration.data
        else:
            self.config = configuration
        self.available_hex_converters = available_hex_converters
        self.procedure_fingerprint_directory = procedure_fingerprint_directory
        self.file_type_dir = file_type_dir
        if plotting_available:
            self.plot = plot
        else:
            self.plot = False
        self.verbose = verbose
        self.timeout = timeout
        # set default exe dir, when none given
        if (
            seabird_exe_directory is None
            or not Path(seabird_exe_directory).exists()
        ):
            self.exe_directory = default_seabird_exe_path()
        else:
            self.exe_directory = Path(seabird_exe_directory)
        self.xmlcon = None
        if auto_run:
            self.output = self.run()

    def run(
        self,
        file: CTDData | CnvFile | HexFile | Path | str = "",
    ) -> CTDData:
        """
        Runs given file or uses the one inside of the config.

        A 'run' consists of the application of all the given modules to a
        given file. It is the structure that can be represented by a
        fingerprint file.

        Parameters
        ----------
        files: CnvFile | Path | str :
            The input file.
             (Default value = '')
        """
        try:
            input = self.config["input"]
        except KeyError:
            input = ""
        except TypeError as error:
            raise IncompleteConfigFile(
                f"Input information has a wrong format: {error}. Aborting."
            )
        file = file if file else input
        if isinstance(file, CnvFile):
            self.ctd_data = file.to_ctd_data()
            self.target = file.path_to_file
            self.load_config()
        elif isinstance(file, HexFile):
            self.target = file.path_to_file
            self.load_config()
            self.ctd_data = decode_hex(file, downcast_only=self.downcast_only)
        elif isinstance(file, Path | str):
            self.target = Path(input) if file == "" else Path(file)
            if self.target == Path("."):
                raise ValueError(
                    "Wrong target file input! \n"
                    + f"Configuration given: {self.config} and file parameter: {file}"
                )
            self.load_config()
            # check whether we are working on hexes or cnvs and react accordingly
            first_module = list(self.modules.keys())[0]
            if first_module.lower() in self.available_hex_converters:
                # remove the first module information and only use it for the
                # conversion.
                hex_converter = self.modules.pop(first_module)
                # convert files to cnvs and continue working as usual
                if first_module.lower() == "datcnv":
                    self.ctd_data = self.convert(
                        self.target, {first_module: hex_converter}
                    )
                elif first_module.lower() == "hex2py":
                    self.ctd_data = decode_hex(
                        self.target,
                        downcast_only=self.downcast_only,
                        **hex_converter,
                    )
            else:
                # auto-convert if hex given as input
                if self.target.suffix == ".hex":
                    self.ctd_data = decode_hex(
                        self.target, downcast_only=self.downcast_only
                    )
                else:
                    self.ctd_data = CnvFile(self.target).to_ctd_data()
        else:
            self.ctd_data = file
            self.target = file.metadata_source.path_to_file
            self.load_config()
        return_value = self.go()
        self.procedure_fingerprint()
        if self.plot:
            basic_bokeh_plot(return_value, print_plot=True, show_plot=False)

        return return_value

    def load_config(self):
        """
        Thorough input/format check of the processing configuration, that
        either stems from a .toml config file, or is a self-build dictionary.
        Checks for the presence of certain keys, and then, depending on their
        importance, either fails or sets default values.
        """
        self.psa_directory = self.check_config_entry("psa_directory", None)
        self.output_dir = Path(
            self.check_config_entry("output_dir", self.target.parent)
        )
        if not self.output_dir.exists():
            self.output_dir.mkdir(parents=True)
        self.output_name = self.check_config_entry("output_name", None)
        self.output_type = self.check_config_entry("output_type", "cnv")
        self.output_columns = self.check_config_entry("output_columns", "all")
        self.xmlcon = self.check_config_entry("xmlcon", None)
        self.bad_flag = self.check_config_entry("bad_flag", -9.990e-29)
        self.downcast_only = self.check_config_entry("downcast_only", True)
        self.remove_flags = self.check_config_entry("remove_flags", True)
        try:
            self.modules = self.config["modules"].copy()
            assert isinstance(self.modules, dict)
            for module in list(self.modules.values()):
                assert isinstance(module, dict)
        except KeyError:
            raise IncompleteConfigFile(
                "No processing modules given. Aborting."
            )
        except AssertionError as error:
            raise IncompleteConfigFile(
                f"Module information is misconfigured: {error}. Aborting."
            )

    def check_config_entry(self, key: str, default_value):
        """Handles configuration file entries."""
        try:
            value = self.config[key]
        except KeyError:
            value = default_value
        else:
            if len(str(value)) == 0:
                value = default_value
        return value

    def is_seabird_module(self, module: dict) -> bool:
        """
        Answers the simple boolean question, whether the module in question is
        a seabird module or not.

        Does that by checking for the presence of a certain key 'psa'. All
        modules that are meant to run as a standalone executable should follow
        this principle and set their config file to the psa key.

        Parameters
        ----------
        module: dict :
            The specific module parameters.

        Returns
        -------

        """
        for key in list(module.keys()):
            if key == "psa":
                return True
        return False

    def create_seabird_module(self, module_info: dict) -> ProcessingModule:
        module_name = list(module_info.keys())[0]
        psa = Path(module_info[module_name]["psa"])
        if len(psa.parents) > 0 and not self.psa_directory:
            self.psa_directory = psa.parent
        module = ProcessingModule(
            name=module_name,
            exe_dir=self.exe_directory,
            psa_dir=self.psa_directory,
            psa_path=psa,
        )
        if "file_suffix" in module_info[module_name].keys():
            module.new_file_suffix = module_info[module_name]["file_suffix"]
        return module

    def create_seabird_step(
        self,
        module: ProcessingModule,
        input_path: Path,
        output_name: Path | str | None = None,
    ) -> ProcessingStep:
        output_name = output_name if output_name else input_path.name
        if input_path.suffix == ".hex":
            hex_file = HexFile(input_path)
            if hex_file.xmlcon:
                self.xmlcon = hex_file.xmlcon.path_to_file.absolute()
        step = ProcessingStep(
            module=module,
            input_path=input_path,
            xmlcon_path=self.xmlcon,
            output_path=self.output_dir,
            original_input_path=self.target,
            new_name=str(output_name),
            verbose=self.verbose,
        )
        return step

    def convert(
        self,
        hex_path: Path,
        hex_converter: dict,
    ) -> CTDData:
        """
        Covers the conversion of hex to cnv file.

        At the moment, this is simply done by using DatCnv, so we could just
        use the general Sea-Bird module pipeline. This is therefore meant to
        be future-compatible for a time where we might have developed other
        hex-converters.

        Parameters
        ----------
        hex_path: Path :
            The path to the target hex file.

        hex_converter: dict :
            The module parameters for the conversion.

        Returns
        -------

        """
        module = self.create_seabird_module(hex_converter)
        output_file_path = self.new_file_path(hex_path.with_suffix(".cnv"))
        step = self.create_seabird_step(module, hex_path, output_file_path)
        step.run()
        try:
            cnv = CnvFile(output_file_path)
        except FileNotFoundError:
            message = f"Failed to convert {hex_path}, using {list(hex_converter.keys())[0]}."
            sys.exit(message)
        else:
            self.xmlcon = step.xmlcon
            return cnv.to_ctd_data()

    def new_file_path(self, file: Path = Path(".")) -> Path:
        """
        Creates the new output file path.

        Takes the file type directory or the given output directory and joins
        them with the given output name.

        Parameters
        ----------
        file: Path :
            The current path to the target file.

        Returns
        -------
        The new path to write the target file to.

        """
        new_name = self.output_name if self.output_name else file.name
        directory = self.output_dir
        return directory.joinpath(new_name)

    def go(self) -> CTDData:
        """
        Performs the processing on all target files.

        This is the 'main' method of the procedure. All previous methods
        prepare data for this method to then finally transform the input files
        into the wanted format.
        The main purpose of this method is the coordination of the two
        different forms of processing modules: standalone executables with
        config files, mainly Sea-Bird processing modules, and python-internal
        classes that implement the Module interface. The caveats are mainly
        the switching from one form to the other. This for example results in
        a in or out parsing of a CnvFile object.

        The output is controlled by self.output_type and is either a cnv file
        at a target path or a CnvFile object.
        """
        path_to_cnv = self.new_file_path(self.target)
        if self.target.suffix == ".cnv":
            try:
                shutil.copy(self.target, path_to_cnv)
            except shutil.SameFileError:
                pass
        last_step = "first"
        for module_name, module_info in self.modules.items():
            module = {module_name.lower(): module_info}
            if self.is_seabird_module(module_info):
                seabird_module = self.create_seabird_module(module)
                step = self.create_seabird_step(
                    module=seabird_module,
                    input_path=path_to_cnv,
                )
                if last_step == "internal":
                    self.ctd_data.to_cnv(file_path=path_to_cnv)
                step.run(timeout=self.timeout)
                last_step = "seabird"
            else:
                if module_name == "hex2py":
                    continue
                own_module = map_proc_name_to_class(module_name)
                if last_step == "seabird":
                    self.ctd_data = CnvFile(path_to_cnv).to_ctd_data()
                try:
                    self.ctd_data = own_module(
                        input=self.ctd_data,
                        arguments=module_info,
                        output_name=str(path_to_cnv),
                        original_input_path=self.target,
                        bad_flag=self.bad_flag,
                    )
                except Exception as error:
                    logger.error(
                        f"Processing step {own_module} failed: {error}. Aborting processing of file {self.ctd_data.path_to_file}."
                    )
                    return self.ctd_data
                if not isinstance(self.ctd_data, CTDData):
                    try:
                        self.btl = self.ctd_data
                        self.ctd_data = self.btl.ctd_data
                    except Exception:
                        logger.error(
                            f"Failed to retrieve new cnv version after step {own_module}. Aborting processing."
                        )
                        break
                last_step = "internal"
        # handle output
        if self.output_type == "cnv":
            if last_step == "internal":
                self.ctd_data.to_cnv(
                    file_path=path_to_cnv,
                    remove_flags=self.remove_flags,
                    output_parameters=self.output_columns,
                    bad_flag=self.bad_flag,
                )
            if self.file_type_dir:
                self.file_type_dir = Path(self.file_type_dir)
                self.handle_fill_type_dir_creation()
        return self.ctd_data

    def procedure_fingerprint(self) -> Configuration | None:
        """
        Handles the creation of individual processing procedure fingerprints.

        A fingerprint is a 'receipt' of one invocation of this class to one or
        more target files or data. It shall serve the purpose of an easy to
        understand proof of what exactly has been done with the data to
        retreive the given result. The especially neat thing is, that they are
        at the very same time plain configuration files, allowing the easy
        re-running of processing procedures. What distinguishes them from the
        usual configuration files, is one, the exact target file list, that can
        be ommitted upon invocation, and a timestamp that prefixes the source
        file name and adheres to ISO 8601.
        """
        if not self.procedure_fingerprint_directory:
            return
        # producing fingerprint that adheres to ISO 8601
        timestamp = datetime.now(timezone.utc).strftime("%Y%m%dT%H%M%SZ")
        directory = Path(self.procedure_fingerprint_directory).absolute()
        if not directory.exists():
            directory.mkdir(parents=True)
        if isinstance(self.config, dict):
            name = Path("processing_config.toml")
            config = Configuration(
                path=directory.joinpath(f"{timestamp}_{name}"),
                data=self.config,
            )
        else:
            config = self.config
            name = config.path.name
        config.write(directory.joinpath(f"{timestamp}_{name}"))
        return config

    def handle_fill_type_dir_creation(self):
        # handle input files, eg. hex, xmlcon, bl
        input_dir = self.target.parent
        file_name = self.target.stem
        for target_file in input_dir.iterdir():
            if target_file.stem == file_name:
                fill_file_type_dir(self.file_type_dir, target_file)
        # handle xmlcon, which can reside in a different directory
        if self.xmlcon:
            fill_file_type_dir(self.file_type_dir, self.xmlcon)
        # handle output files, eg. cnv, btl
        for target_file in self.output_dir.iterdir():
            if target_file.stem.startswith(file_name):
                fill_file_type_dir(self.file_type_dir, target_file)



import logging
import warnings
from copy import copy
from pathlib import Path
from typing import Dict, Tuple

import numpy as np
from scipy.ndimage import convolve1d
from scipy.signal import butter, correlate, filtfilt, find_peaks
from scipy.signal.windows import boxcar, triang
from seabirdscientific import processing as sbs_proc

from ctdam.parser import CnvFile, CTDData, Parameter
from ctdam.proc.module import ArrayModule, MissingParameterError

logger = logging.getLogger(__name__)


class LoopRemoval(ArrayModule):
    def __init__(self) -> None:
        super().__init__()

    def __call__(
        self,
        input: Path | str | CnvFile | CTDData,
        arguments: dict = {},
        output: str = "cnvobject",
        output_name: str | None = None,
        default_values: dict = {
            "precut_period": 5,
            "cut_period": 10,
            "mean_speed_percent": 30,
            "delay": 2,
            "filter_order": 4,
        },
        **kwargs,
    ) -> None | CnvFile | CTDData:
        return super().__call__(
            input, arguments, output, output_name, default_values
        )

    def transformation(self) -> bool:
        """
        Calls the loop removal function and handles the resulting flag values
        for array truncation.
        """
        if not self._check_parameter_existence("prDM"):
            logger.error("Failed, not finding pressure")
            return False

        pressure = self.ctd_data["prDM"].data

        new_flag_array = self.jens_loop_removal(
            pressure=pressure,
            sample_interval=1 / self.sample_rate,
            **self.arguments,
        )

        self.handle_new_flags(new_flag_array)
        return True

    def jens_loop_removal(
        self,
        pressure: np.ndarray,
        sample_interval: float,
        precut_period: int = 5,
        cut_period: int = 10,
        mean_speed_percent: int = 20,
        delay: int = 2,
        filter_order: int = 4,
    ):
        """
        Flag loops in CTD data caused by ship heave.
        Credit: Dr. Jens Faber, IOW.

        Parameters:
        - pressure: Array of vertical axis values (e.g., pressure).
        - time: Array of time values.
        - precut_period: Cutoff period for the pre-filter (seconds).
        - cut_period: Cutoff period for the main filter (seconds).
        - mean_speed_percent: Percentage of filtered velocity to use as a threshold.
        - delay: Delay (in seconds) to shift the flag array.
        - filter_order: Order of the Butterworth filter.

        Returns:
        - flag_bool: Boolean array where `True` indicates a flagged (bad) data point.
        """
        warnings.warn(
            "LoopRemoval is still in an experimental state. Be cautious with the results."
        )
        # Compute vertical velocity
        velocity = np.gradient(pressure) / sample_interval

        # Pre-filtering: Low-pass Butterworth filter
        b, a = butter(
            filter_order, 2 * sample_interval / precut_period, btype="low"
        )
        # Pad the signal
        velocity_padded = np.pad(velocity, (3, 3), mode="edge")
        velocity_filt_pre = filtfilt(b, a, velocity_padded)
        # Remove padding
        velocity_filt_pre = velocity_filt_pre[3:-3]

        # Main filtering: Low-pass Butterworth filter
        b, a = butter(
            filter_order, 2 * sample_interval / cut_period, btype="low"
        )
        # Pad the signal
        velocity_padded = np.pad(velocity_filt_pre, (3, 3), mode="edge")
        velocity_filt = filtfilt(b, a, velocity_padded)
        # Remove padding
        velocity_filt = velocity_filt[3:-3]

        # Flag data where velocity is below the threshold
        flag_bool = velocity < (velocity_filt * mean_speed_percent / 100)

        # Shift the flag array to account for delay
        sample_shift = int(round(delay / sample_interval))
        flag_bool = np.roll(flag_bool, sample_shift)
        # Ensure no flags are set before the delay
        flag_bool[:sample_shift] = False

        return flag_bool


class AlignCTD(ArrayModule):
    """
    Align the given parameter columns.

    Given a measurement parameter in parameters, the column will be shifted
    by either, a float amount that is given as value, or, by a calculated
    amount, using cross-correlation between the high-frequency components of
    the temperature and the target parameters.
    The returned numpy array will thus feature the complete CnvFile data,
    with the columns shifted to their correct positions.
    """

    def __init__(self) -> None:
        super().__init__()

    def __call__(
        self,
        input: Path | str | CnvFile | CTDData,
        arguments: dict = {},
        output: str = "cnvobject",
        output_name: str | None = None,
        default_values: dict = {
            "Oxygen": 3,
            "minimum_correlation": 0.1,
            "default_shift": 3,
        },
        **kwargs,
    ) -> None | CnvFile | CTDData:
        return super().__call__(
            input, arguments, output, output_name, default_values
        )

    def transformation(self) -> bool:
        """
        Performs the base logic of distinguishing whether to use given values
        or compute a delay.

        Returns
        -------
        A numpy array, representing the cnv data after the alignment.

        """
        self.check_whether_working_on_binned_data()
        return_value = False
        new_parameter_metadata = {}
        for key, value in self.handle_parameter_input(self.arguments).items():
            # key is something like oxygen1 or oxygen2
            # value is either None or a numerical value in string or other form
            target_parameters = [
                param
                for param in self.ctd_data.get_parameter_list()
                if (param.param.lower().startswith(key[:-1]))
                and (str(int(key[-1]) - 1) in param.name)
            ]
            # if there are no measurement parameters of the given key inside
            # the cnv file, remove the key from the input, to avoid printing
            # that key to the output files header
            if len(target_parameters) == 0:
                continue
            # if no shift value given, estimate it
            if not value:
                value, correlation_value = self.estimate_sensor_delay(
                    delayed_parameter=target_parameters[0],
                    margin=len(self.ctd_data.get_full_data_array()) // 4,
                )
                correlation_string = f", with PCC: {correlation_value}"
                if not self.check_correlation_result(
                    value,
                    correlation_value,
                    self.arguments["minimum_correlation"],
                ):
                    correlation_string = f", default value. Calculated delay: {str(float('{:.2f}'.format(value / self.sample_rate)))} PCC: {correlation_value}"
                    # set to a default value
                    value = self.arguments["default_shift"] * self.sample_rate
            else:
                # the input is in seconds, so we calculate a shift in rows
                value = float(value) * self.sample_rate
                correlation_string = ""

            if value > self.ctd_data.get_data_length():
                warnings.warn(
                    f"Data size of {self.ctd_data.get_data_length()} too small for shift of {value}. Skipping AlignCTD.",
                    category=RuntimeWarning,
                )
                return False

            # apply shift for all columns of the given parameter
            for parameter in target_parameters:
                # get the number of decimals to format the output in the same
                # way
                number_of_decimals = len(str(parameter.data[0]).split(".")[1])
                # do the shifting/alignment
                parameter.data = np.append(
                    parameter.data[int(value) :,].round(
                        decimals=number_of_decimals
                    ),
                    np.full((int(value),), self.bad_flag),
                )
                # format the output back to seconds
                new_parameter_metadata[parameter.name] = (
                    str(float("{:.2f}".format(value / self.sample_rate)))
                    + "s"
                    + correlation_string
                )
                try:
                    self.array = self.ctd_data.get_full_data_array()
                except IndexError as error:
                    logger.error(
                        f"AlignCTD failed for {self.ctd_data.path_to_file} while aligning {parameter}: {error}"
                    )
                    return_value = False
                    break
                # at least one column has been altered so we can give positive
                # feedback
                return_value = True
        self.arguments = new_parameter_metadata
        return return_value

    def estimate_sensor_delay(
        self,
        delayed_parameter: Parameter,
        margin: int = 240,
        shift_seconds: int = 10,
    ) -> Tuple[float, float]:
        """
        Estimate delay between a delayed parameter and temperature signals via
        cross-correlation of high-frequency components.

        Parameters
        ----------
        delayed_parameter: Parameter :
            The parameter whose delay shall be computed.

        margin: int :
            A number of data points that are cutoff from both ends.
             (Default value = 240)

        shift_seconds: int :
             Maximum time window to search for lag (default: 10 seconds).

        Returns
        -------
        A float value, representing the parameter delay in seconds.

        """
        temperature = self.find_corresponding_temperature(
            delayed_parameter
        ).data
        delayed_values = delayed_parameter.data
        assert len(temperature) == len(delayed_values)
        # remove edge effects (copying Gerds MATLAB software)
        while len(temperature) <= 2 * margin:
            margin = margin // 2

        t_shortened = np.array(temperature[margin:-margin])
        v_shortened = np.array(delayed_values[margin:-margin])

        if np.all(np.isnan(v_shortened)):
            return np.nan, np.nan

        # design Butterworth filter
        b, a = butter(3, 0.005)

        # smooth signals
        t_smoothed = filtfilt(b, a, t_shortened)
        v_smoothed = filtfilt(b, a, v_shortened)

        # high-frequency components
        t_high_freq = t_shortened - t_smoothed
        v_high_freq = v_shortened - v_smoothed

        # cross-correlation
        max_lag = int(shift_seconds * self.sample_rate)
        sign = self.get_correlation(delayed_parameter)
        corr = correlate(v_high_freq, t_high_freq * sign, mode="full")
        lags = np.arange(-len(t_high_freq) + 1, len(t_high_freq))
        lag_indices = np.where(np.abs(lags) <= max_lag)[0]

        # normalize correlation values
        norm_factor = np.sqrt(np.sum(v_high_freq**2) * np.sum(t_high_freq**2))
        corr_normalized = corr / norm_factor

        corr_segment = corr_normalized[lag_indices]
        lags_segment = lags[lag_indices]

        # restrict to only positive delays
        positive_indices = np.where(lags_segment > 0)[0]
        corr_segment_positive = corr_segment[positive_indices]

        peaks, props = find_peaks(
            corr_segment_positive, height=0.01, distance=5
        )

        # handle case, when no correlation can be found
        if len(peaks) == 0:
            return np.nan, np.nan

        # find lag with highest correlation
        best_index = int(np.argmax(props["peak_heights"]))

        return float(peaks[best_index]), float(
            "{:.2f}".format(props["peak_heights"][best_index])
        )

    def check_correlation_result(
        self,
        value: float,
        correlation_value: float,
        minimum_correlation: float = 0.1,
    ) -> bool:
        """
        Performs several checks on the delay outputed by
        self.estimate_sensor_delay and returns True, if the result is
        considered feasible.
        """
        if (value is np.nan) or (correlation_value is np.nan):
            return False
        value = value / self.sample_rate
        if correlation_value < minimum_correlation:
            return False
        if value < 1 or value > 6:
            return False
        return True

    def find_corresponding_temperature(
        self, parameter: Parameter
    ) -> Parameter:
        """
        Find the temperature values of the sensor that shared the same water
        mass as the input parameter.

        Parameters
        ----------
        parameter: Parameter :
            The parameter of interest.


        Returns
        -------
        The temperature parameter object.

        """
        if "0" in parameter.name:
            return self.ctd_data["t090C"]
        elif "1" in parameter.name:
            return self.ctd_data["t190C"]
        else:
            raise MissingParameterError("AlignCTD", "Temperature")

    def get_correlation(self, parameter: Parameter) -> float:
        """
        Gives a number indicating the cross correlation type regarding the
        input parameter and the temperature.

        Basically distinguishes between positive correlation, 1, and anti-
        correlation, -1. This value is then used to alter the temperature
        values accordingly.

        Parameters
        ----------
        parameter: Parameter :
            The parameter to cross correlate with temperature.

        Returns
        -------
        A float value representing positive or negative correlation.

        """
        if parameter.metadata["name"].lower().startswith("oxygen"):
            return -1
        else:
            return 1

    def handle_parameter_input(self, input_dict: dict) -> dict:
        new_dict = {}
        all_parameter_names = [
            value["name"].lower()
            for value in self.ctd_data.get_metadata().values()
        ]
        for parameter_input, value in input_dict.items():
            # remove all non-alphanumeric characters
            parameter = (
                "".join(filter(str.isalnum, parameter_input)).lower().strip()
            )
            if parameter_input[-1] in ["1", "2"]:
                parameter = parameter[:-1]
                number = parameter_input[-1]
            else:
                number = None
            parameter_names = [
                name
                for name in all_parameter_names
                if name.startswith(parameter)
            ]
            # check, whether we are working with multiple sensors
            if "2" in [name[-1] for name in parameter_names]:
                # differentiate the different cases for 2 sensors
                # only parameter without sensor number information given
                if parameter.lower() in parameter_names and not number:
                    new_dict[f"{parameter}1"] = value
                    new_dict[f"{parameter}2"] = value
                # explicitly given sensor 1
                if parameter.lower() in parameter_names and number == "1":
                    new_dict[f"{parameter}1"] = value
                # explicitly given sensor 2
                if parameter.lower() in parameter_names and number == "2":
                    new_dict[f"{parameter}2"] = value
            else:
                # single sensor is easy, just use the value for sensor 1
                if not parameter[-1] == "2":
                    new_dict[f"{parameter}1"] = value
        return new_dict


class WFilter(ArrayModule):
    def __init__(self) -> None:
        super().__init__()

    def __call__(
        self,
        input: Path | str | CnvFile | CTDData,
        arguments: dict = {},
        output: str = "cnvobject",
        output_name: str | None = None,
        default_values: dict = {
            "Pressure": {
                "window_type": "gaussian",
                "window_width": 20,
                "half_width": 0.415,
                "offset": 0,
            },
            "Temperature": {
                "window_type": "gaussian",
                "window_width": 24,
                "half_width": 0.5,
                "offset": 0,
            },
            "Conductivity": {
                "window_type": "gaussian",
                "window_width": 24,
                "half_width": 0.5,
                "offset": 0,
            },
            "Salinity": {
                "window_type": "gaussian",
                "window_width": 24,
                "half_width": 0.5,
                "offset": 0,
            },
            "Oxygen": {
                "window_type": "gaussian",
                "window_width": 48,
                "half_width": 1,
                "offset": 0,
            },
            "Fluorescence": {
                "window_type": "median",
                "window_width": 5,
                "half_width": 1,
                "offset": 0,
            },
            "Turbidity": {
                "window_type": "median",
                "window_width": 5,
                "half_width": 1,
                "offset": 0,
            },
            "PAR": {
                "window_type": "median",
                "window_width": 5,
                "half_width": 1,
                "offset": 0,
            },
            "SPAR": {
                "window_type": "median",
                "window_width": 5,
                "half_width": 1,
                "offset": 0,
            },
            "FlowMeter": {
                "window_type": "median",
                "window_width": 5,
                "half_width": 1,
                "offset": 0,
            },
        },
        **kwargs,
    ) -> None | CnvFile | CTDData:
        self.default_values = default_values
        return super().__call__(input, arguments, output, output_name)

    def transformation(self) -> bool:
        """
        Performs the base logic of distinguishing whether to use given values
        or compute a delay.

        Returns
        -------
        A numpy array, representing the cnv data after the alignment.

        """
        general_kwargs = {
            "flags": self.flags,
            "sample_interval": 1 / self.sample_rate,
            "exclude_flags": False,
            "flag_value": self.bad_flag,
        }
        # sanitize user input
        iter_arguments = copy(self.arguments)
        for key, value in iter_arguments.items():
            self.arguments[key.replace(" ", "").lower()] = value
            self.arguments.pop(key)
        new_arguments = {}
        for param in self.ctd_data.parameters.get_parameter_list():
            try:
                specific_kwargs = self.default_values[param.param]
            except KeyError:
                specific_kwargs = {}
            if param.param.lower() in self.arguments:
                # use default values of SPAR, to allow the user to not set all
                # 4 values that are necessary to run a wfilter
                specific_kwargs = self.default_values["SPAR"]
                for key, value in self.arguments[param.param.lower()].items():
                    if key == "window_type":
                        value = value.lower()
                    specific_kwargs[key] = value
            if specific_kwargs:
                with warnings.catch_warnings(action="ignore"):
                    param.data = self.window_filter(
                        data_in=param.data,
                        **general_kwargs,
                        **specific_kwargs,
                    )
                new_arguments[param.param] = ", ".join(
                    [str(value) for value in specific_kwargs.values()]
                )

        self.arguments = new_arguments

        return True

    def window_filter(
        self,
        data_in: np.ndarray,
        flags: np.ndarray,
        window_type: str,
        window_width: int,
        sample_interval: float,
        half_width: float = 1.0,
        offset: float = 0.0,
        exclude_flags: bool = False,
        flag_value: float = -9.99e-29,
    ) -> np.ndarray:
        """Filters a dataset by convolving it with an array of weights.

        The available window filter types are boxcar, cosine, triangle,
        gaussian, and median. Refer to the SeaSoft data processing manual
        version 7.26.8, page 108.

        Args:
            data_in: Data to be filtered.
            flags: Flagged data defined by loop edit.
            window_type: The filter type (boxcar, cosine, triangle, gaussian, or median).
            window_width: Width of the window filter (must be odd).
            sample_interval: Sample interval of the dataset.
            half_width: Width of the Gaussian curve.
            offset: Shifts the center point of the Gaussian.
            exclude_flags: Exclude flagged values from the dataset.
            flag_value: The flag value in flags.

        Returns:
            The convolution of data_in and the window filter.
        """
        # Convert flags to NaN for processing
        data = np.where(data_in == flag_value, np.nan, data_in)
        if exclude_flags:
            data = np.where(flags == flag_value, np.nan, data)

        # Define the window filter
        window_start = -(window_width - 1) // 2
        window_end = (window_width - 1) // 2 + 1

        if window_type == "boxcar":
            window = boxcar(window_width)
        elif window_type == "cosine":
            n = np.arange(window_start, window_end)
            window = np.cos((n * np.pi) / (window_width + 1))
        elif window_type == "triangle":
            window = triang(window_width)
        elif window_type == "gaussian":
            phase = offset / sample_interval
            scale = np.log(2) * (2 * sample_interval / half_width) ** 2
            n = np.arange(window_start, window_end)
            window = np.exp(-((n - phase) ** 2) * scale)
        elif window_type == "median":
            pass
        else:
            logger.warning(
                f"No known window_type: {window_type}. Skipping wfilter."
            )
            return data

        padding_size = window_width // 2

        # Pad data for convolution
        data_valid = np.nan_to_num(data)
        data_padded = np.pad(
            data_valid,
            padding_size,
            mode="edge",
        )

        # Handle NaN values: replace with 0 for convolution, then mask later
        nan_mask = np.isnan(data_padded)
        data_filled = np.where(nan_mask, 0, data_padded)

        # Convolve using SciPy's convolve1d (handles edge cases better)
        if window_type == "median":
            # For median, use a sliding window approach (no direct vectorization)
            data_out = np.array(
                [
                    np.nanmedian(data_padded[i : i + window_width])
                    for i in range(len(data))
                ]
            )
        else:
            # Normalize the window
            window_normalized = window / np.sum(window)

            # Convolve
            conv_result = convolve1d(
                data_filled,
                window_normalized,
                mode="constant",
                origin=-(window_width // 2),
            )

            # Restore NaN values where they were in the original data
            data_out = np.where(
                nan_mask[padding_size:-padding_size],
                np.nan,
                conv_result[padding_size:-padding_size],
            )

        return data_out


class CellTM(ArrayModule):
    def __init__(self) -> None:
        super().__init__()

    def __call__(
        self,
        input: Path | str | CnvFile | CTDData,
        arguments: dict = {},
        output: str = "cnvobject",
        output_name: str | None = None,
        default_mapping: dict = {
            "sbe9": (0.03, 7.0),
            "sbe19": (0.04, 8.0),
        },
        **kwargs,
    ) -> None | CnvFile | CTDData:
        self.cell_tm_param_mapping = default_mapping
        return super().__call__(input, arguments, output, output_name)

    def transformation(self) -> bool:
        if "alpha" in self.arguments and "beta" in self.arguments:
            self.alpha = self.arguments["alpha"]
            self.beta = self.arguments["beta"]
        else:
            try:
                for key in self.cell_tm_param_mapping:
                    if key in self.ctd_data.header[0].lower().replace(" ", ""):
                        self.alpha, self.beta = self.cell_tm_param_mapping[key]

            except KeyError:
                logger.error(
                    f"No cell_tm parameters for instrument {self.ctd_data.header[0][:-10]}. No cell thermal mass correction applied."
                )
            else:
                self.arguments["alpha"] = self.alpha
                self.arguments["beta"] = self.beta
        for param in [p for p in self.ctd_data if p.param == "Conductivity"]:
            # check availability of temperature in this sensor strand
            if param.sensor_number == 1:
                temperature_name = "t090C"
            else:
                temperature_name = "t190C"
            temperature = self.ctd_data[temperature_name].data
            if not self._check_parameter_existence(temperature_name):
                logger.error(
                    f"Missing temperature for sensor strand {param.sensor_number}"
                )
                return False

            # enforce correct conductivity unit
            if param.unit == "mS/cm":
                conductivity = param.data / 10.0
            elif param.unit == "S/m":
                conductivity = param.data
            else:
                logger.error(
                    f"Unknown conductivity unit {param.unit}. Aborting."
                )
                return False
            # seabirds celltm cannot handle nans, setting so bad flag value
            temperature = np.nan_to_num(temperature, nan=self.bad_flag)
            param.data[param.data == self.bad_flag] = np.nan
            corrected_conductivity = sbs_proc.cell_thermal_mass(
                temperature_C=temperature,
                conductivity_Sm=conductivity,
                amplitude=self.alpha,
                time_constant=1 / self.beta,
                sample_interval=1 / self.sample_rate,
            )

            if param.unit == "mS/cm":
                param.data = corrected_conductivity * 10
            elif param.unit == "S/m":
                param.data = corrected_conductivity

        return True


class BinAvg(ArrayModule):
    def __init__(self) -> None:
        super().__init__()

    def __call__(
        self,
        input: Path | str | CnvFile | CTDData,
        arguments: dict = {},
        output: str = "cnvobject",
        output_name: str | None = None,
        default_values: dict = {
            "bin_variable": "prDM",
            "bin_size": 1,
            "cast_type": "down",
        },
        **kwargs,
    ) -> None | CnvFile | CTDData:
        self.name = "binning"
        return super().__call__(
            input, arguments, output, output_name, default_values
        )

    def transformation(self) -> bool:
        """
        Calls the loop removal function and handles the resulting flag values
        for array truncation.
        """
        self.check_whether_working_on_binned_data()
        self.ctd_data.drop_flagged_rows()
        for param in self.ctd_data:
            param.data = np.nan_to_num(param.data, nan=self.bad_flag)

        dataset = {param.name: param.data for param in self.ctd_data}
        try:
            array_data = self.own_bin_average(
                data=dataset,
                flag_value=self.bad_flag,
                **self.arguments,
            )
        except Exception as error:
            logger.error(
                f"Could not bin {self.ctd_data.path_to_file}: {error}"
            )
            return False
        for name, data in array_data.items():
            for param in self.ctd_data:
                if param.name == name:
                    param.data = data

        self.ctd_data[self.arguments["bin_variable"]].data = np.round(
            self.ctd_data[self.arguments["bin_variable"]].data, 0
        )
        # set new sample rate
        if self.arguments["bin_variable"] == "prDM":
            bin_unit = "decibars"
        elif self.arguments["bin_variable"] == "timeS":
            bin_unit = "seconds"
        else:
            bin_unit = self.arguments["bin_variable"]

        self.ctd_data.set_sample_rate(
            float(self.arguments["bin_size"]), bin_unit
        )

        return True

    def own_bin_average(
        self,
        data: Dict[str, np.ndarray],
        bin_variable: str,
        bin_size: float,
        min_scans: int = 1,
        max_scans: int = 999999,
        cast_type: str = "down",
        flag_value: float = -9.99e-29,
        include_scan_count: bool = True,
    ) -> Dict[str, np.ndarray]:
        """
        Optimized bin average using a vectorized approach on numpy arrays.
        """

        # Validate and filter bad data
        n_rows = len(data[bin_variable])
        valid_mask = np.ones(n_rows, dtype=bool)
        for arr in data.values():
            valid_mask &= arr != flag_value

        control = data[bin_variable][valid_mask]
        filtered_data = {col: arr[valid_mask] for col, arr in data.items()}
        n_valid = len(control)

        if n_valid == 0:
            return {col: np.array([]) for col in data.keys()}

        peak_idx = np.nanargmax(control)

        # Setup bins
        bin_min = 0 if bin_variable == "nScan" else bin_size / 2.0
        control_max = np.max(control)
        bin_max = control_max - ((bin_min + control_max) % bin_size) + bin_size
        bin_edges = np.arange(bin_min, bin_max + bin_size, bin_size)

        # Assign bin numbers
        desc_control = control[: peak_idx + 1]
        asc_control = control[peak_idx:]

        desc_bins = np.digitize(desc_control, bin_edges)
        asc_bin_edges = np.arange(bin_max, bin_min - bin_size, -bin_size)
        asc_bins = np.digitize(asc_control, asc_bin_edges, right=True)
        asc_bins += np.nanmax(desc_bins) - 1

        bin_numbers = np.concatenate([desc_bins[:-1], asc_bins])

        # Apply cast type filter
        if cast_type == "down":
            keep_mask = (np.arange(n_valid) <= peak_idx) & (bin_numbers >= 1)
        elif cast_type == "up":
            min_bin = np.min(asc_bins) + 1
            max_bin = np.max(asc_bins) - 1
            keep_mask = (
                (np.arange(n_valid) >= peak_idx)
                & (bin_numbers >= min_bin)
                & (bin_numbers <= max_bin)
            )
        elif cast_type == "both":
            max_bin = np.max(asc_bins) - 1
            keep_mask = (bin_numbers >= 1) & (bin_numbers <= max_bin)
        else:
            keep_mask = np.ones(n_valid, dtype=bool)

        bin_numbers = bin_numbers[keep_mask]
        filtered_data = {
            col: arr[keep_mask] for col, arr in filtered_data.items()
        }

        if len(bin_numbers) == 0:
            return {col: np.array([]) for col in data.keys()}

        # Use bincount for fast averaging
        max_bin_num = bin_numbers.max()
        bin_counts = np.bincount(bin_numbers, minlength=max_bin_num + 1)

        # Find valid bins
        unique_bins = np.where(
            (bin_counts >= min_scans)
            & (bin_counts <= max_scans)
            & (bin_counts > 0)
        )[0]

        if len(unique_bins) == 0:
            return {col: np.array([]) for col in data.keys()}

        # Compute averages using bincount (much faster!)
        results = {}
        for col_name, arr in filtered_data.items():
            if col_name == "flag":
                binned_values = np.zeros(len(unique_bins))
                for i, bin_num in enumerate(unique_bins):
                    bin_mask = bin_numbers == bin_num
                    bin_vals = arr[bin_mask]
                    binned_values[i] = (
                        flag_value if np.all(bin_vals == flag_value) else 0
                    )
                results[col_name] = binned_values
            else:
                bin_sums = np.bincount(
                    bin_numbers, weights=arr, minlength=max_bin_num + 1
                )
                bin_means = bin_sums / bin_counts
                results[col_name] = bin_means[unique_bins]

        if include_scan_count:
            results["nbin"] = bin_counts[unique_bins]

        return results



import logging
from pathlib import Path

import numpy as np
import pandas as pd

from ctdam.parser.ctddata import CTDData
from ctdam.proc.module import ArrayModule

logger = logging.getLogger(__name__)


class AirPressureCorrection(ArrayModule):
    """
    Corrects water pressure by the given air pressure.
    """

    def __call__(
        self,
        input: Path | str | CTDData | pd.DataFrame | np.ndarray,
        arguments: dict = {},
        output: str = "cnvobject",
        output_name: str | None = None,
        **kwargs,
    ) -> None | CTDData | pd.DataFrame | np.ndarray:
        return super().__call__(input, arguments, output, output_name)

    def transformation(self) -> bool:
        """
        Base logic to correct pressure.
        """
        try:
            prDM = self.ctd_data["prDM"].data
            air_pressure = float(
                self.ctd_data.metadata["Air_Pressure"].replace("hPa", "")
            )
        except (KeyError, ValueError):
            return False

        water_pressure = 1024
        pressure_diff = round((air_pressure - water_pressure) / 100, 4)
        self.ctd_data["prDM"].data = prDM - pressure_diff
        self.arguments["pressure_diff"] = f"{str(-pressure_diff)} dbar"

        return True



from pathlib import Path

import pytest
from conftest import (
    btl_path,
    check_and_remove_file,
    cnv_path,
    hex_path,
    proc_template,
    psa_path,
    test_hex,
)
from numpy.testing import assert_equal

from ctdam.parser.ctddata import CTDData
from ctdam.proc.procedure import Procedure
from ctdam.proc.settings import IncompleteConfigFile


@pytest.mark.long
@pytest.mark.parametrize(
    "filename", [filename for filename in hex_path.glob("*.hex")]
)
def test_procedure_on_hexes(filename: Path, tmp_path):
    output_name = f"{filename.stem}_test.cnv"
    proc_config = {
        "input": filename,
        "output_dir": tmp_path,
        "output_name": output_name,
        "output_type": "cnv",
        "modules": {
            "alignctd": {},
            "binavg": {},
        },
    }
    procedure = Procedure(
        proc_config,
        auto_run=False,
        timeout=10,
    )
    procedure.run()
    for suffix in [""]:
        check_and_remove_file(
            tmp_path.joinpath(f"{filename.stem}_test{suffix}.cnv")
        )


def test_fingerprint(tmp_path):
    proc_config = {
        "input": cnv_path,
        "output_type": "internal",
        "modules": {
            "alignctd": {"Oxygen": ""},
        },
    }
    procedure = Procedure(
        proc_config,
        auto_run=False,
        procedure_fingerprint_directory=tmp_path,
    )
    new_config = procedure.procedure_fingerprint()
    if new_config is not None:
        assert proc_config == new_config.data
    else:
        pytest.fail("Expected a configuration from fingerprint.")


def test_procedure_without_seabird(tmp_path):
    cnv_name = "EMB295_14-1"
    file_type_dir = tmp_path.joinpath("procedure")
    proc_config = {
        "input": cnv_path.joinpath(cnv_name).with_suffix(".cnv"),
        "output_type": "cnv",
        "output_dir": tmp_path,
        "modules": {
            "alignctd": {"Oxygen": ""},
            "wildedit_geomar": {
                "std1": "3",
                "std2": 4,
                "window_size": "200",
            },
            "Helmholtz_energy_ice": {},
            "create_bottle_file": {
                "bl": str(btl_path.joinpath(cnv_name).with_suffix(".bl")),
            },
        },
    }
    procedure = Procedure(proc_config, file_type_dir=file_type_dir)
    assert "gsw_Helmholtz_0" in procedure.ctd_data.parameters
    assert (
        "create_bottle_file"
        not in procedure.ctd_data.processing_steps.get_names()
    )
    assert procedure.btl.ctd_data == procedure.ctd_data
    for file in [
        tmp_path.joinpath(cnv_name).with_suffix(".cnv"),
        tmp_path.joinpath(cnv_name).with_suffix(".obtl"),
        file_type_dir.joinpath("obtl", cnv_name).with_suffix(".obtl"),
    ]:
        check_and_remove_file(file)


def test_empty_modules():
    with pytest.raises(IncompleteConfigFile):
        Procedure(
            {
                "input": test_hex,
                "psa_directory": psa_path,
            }
        )


@pytest.mark.long
def test_full_file_proc(tmp_path):
    Procedure(
        proc_template,
        auto_run=False,
        file_type_dir=tmp_path,
        procedure_fingerprint_directory=tmp_path,
    ).run(test_hex)


@pytest.mark.long
@pytest.mark.parametrize(
    "hex", [filename for filename in hex_path.glob("*.hex")]
)
def test_non_seabird_conversion_and_processing(hex, create_files, tmp_path):
    if hex.name.startswith("SO308-2_005"):
        pytest.skip()
    proc_config = {
        "input": hex,
        "output_type": "ctd_data",
        "output_dir": tmp_path,
        "modules": {
            "wildedit_geomar": {},
            "wfilter": {},
            "airpressure": {},
            "celltm": {},
            "alignctd": {},
            "Helmholtz_energy_ice": {},
            "loop_removal": {},
            "binavg": {},
        },
    }
    procedure = Procedure(proc_config, plot=True)
    file_path = procedure.ctd_data.path_to_file
    assert isinstance(procedure.ctd_data, CTDData)
    num_of_proc_steps = len(proc_config["modules"]) + 1
    if "Air_Pressure" not in procedure.ctd_data.metadata:
        num_of_proc_steps -= 1
    assert len(procedure.ctd_data.processing_steps) == num_of_proc_steps
    # check, whether salinity has been recalculated
    if create_files:
        pre_salinity = procedure.ctd_data.parameters["sal00"].data
        procedure.ctd_data.to_cnv(
            f"procedure_{file_path.with_suffix('.cnv')}",
            remove_flags=False,
        )
        try:
            assert_equal(
                pre_salinity, procedure.ctd_data.parameters["sal00"].data
            )
        except AssertionError:
            assert True
        else:
            assert False


def test_conversion_options():
    proc_config = {
        "input": test_hex,
        "output_type": "internal",
        "modules": {
            "hex2py": {
                "min_soak_window": 50,
                "max_fd_quotient": 100,
                "min_velocity": 2.0,
            },
            "binavg": {},
        },
    }
    procedure = Procedure(proc_config)
    assert procedure.output.cast_borders["down_start"] == 0



import difflib
import logging
from pathlib import Path

logger = logging.getLogger(__name__)


class ProcessingModule:
    """
    Collects the information necesssary to run a Seabird Processing module.


    A module needs paths to an exe, to a psa and to an input file. On top of
    that, a few non-mandatory settings can be specified, like the output file
    path or a new file name. The collected information is then used to feed a
    ProcessingStep.

    Parameters
    ----------

    name: str:
        The name of the module. Is used to find corresponding files.

    exe_dir: Path | str:
        The path to the executable directory. Is used if no direct path to any
        executable is given.

    psa_dir: Path | str:
        The path to the configuration directory. Is used if no direct path to
        any config file is given.

    exe_path: Path | str | None:
        The path to the modules executable.

    psa_path: Path | str | None:
        The path to the modules configuration file.

    new_file_suffix: str:
        An optional suffix to append to the output file name.
    """

    def __init__(
        self,
        name: str,
        exe_dir: Path | str,
        psa_dir: Path | str,
        exe_path: Path | str | None = None,
        psa_path: Path | str | None = None,
        new_file_suffix: str = "",
    ):
        # allowing the usage of multiple modules of the same name inside of one
        # routine/procedure
        self.name = name.lower().split("_")[0]
        if self.name == "airpressure":
            self.psa = self.check_input_path(psa_path, "par", Path(psa_dir))
        elif self.name == "iow_btl_id":
            self.psa = None
        # w_filter will be split also and needs to be handled individually
        elif self.name == "w":
            self.name = "w_filter"
            self.psa = self.check_input_path(psa_path, "psa", Path(psa_dir))
        else:
            self.psa = self.check_input_path(psa_path, "psa", Path(psa_dir))
        self.exe = self.check_input_path(exe_path, "exe", Path(exe_dir))
        self.new_file_suffix = new_file_suffix

    def __str__(self) -> str:
        return self.name

    def check_input_path(
        self, input_path: Path | str | None, file_type: str, directory: Path
    ) -> Path | None:
        """
        Tests the given exe or config path and runs find_file upon failure.

        The path needs to be absolute and present. This method basically
        performs all the logic to retrieve an executable and configuration path
        with the help of find_file.

        Parameters
        ----------
        input_path: Path | str | None :
            The path to the file that needs testing. If none given, directly
            start searching a default one by calling find_file.

        file_type: str :
            The type of the file to check on. Either 'exe' or some form of
            configuration file, usually 'psa'.

        directory: Path :
            The directory to search for, upon failure.
        """
        if isinstance(input_path, Path | str):
            input_path = Path(input_path)
        else:
            return self.find_file(file_type, directory)
        if input_path.is_absolute():
            output_path = input_path
        else:
            output_path = directory.joinpath(input_path)
        if output_path.exists():
            return output_path
        else:
            return self.find_file(file_type, directory)

    def find_file(self, file_type: str, parent_dir: Path) -> Path | None:
        """
        Finds the closest matching file of a certain type inside of a given
        directory.

        Parameters
        ----------
        file_type: str :
            The file type to search for.

        parent_dir: Path :
            The directory to search in.
        """
        file_type = file_type.lstrip(".")
        try:
            best_match = difflib.get_close_matches(
                word=f"{self.name}.{file_type}",
                possibilities=[path.name for path in parent_dir.iterdir()],
                n=1,
            )
        except FileNotFoundError as error:
            logger.error(
                f'The {file_type} directory "{parent_dir}" could not be found: {error}'
            )
        else:
            try:
                return parent_dir.joinpath(best_match[0])
            except IndexError as error:
                logger.error(
                    f"No exe or config given and none found using the name {
                        self.name
                    }.{file_type} in {parent_dir}: {error}"
                )
